#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <unistd.h>
#include <libgen.h>

enum data_type {
  LE_STR,
  LE_BIN
};

struct hdr_desc {
  int size;
  data_type type;
  char *desc;
};
  
hdr_desc le_hib_tbl[] = {
  { 2, LE_STR, "Signature" },
  { 1, LE_BIN, "Byte order" },
  { 1, LE_BIN, "Word order" },
  { 4, LE_BIN, "Executable format level" },
  { 2, LE_BIN, "CPU Type" },
  { 2, LE_BIN, "Target operating system" },
  { 4, LE_BIN, "Module version" },
  { 4, LE_BIN, "Module type flags" },
  { 4, LE_BIN, "Number of memory pages" },
  { 4, LE_BIN, "Initial object CS number" },
  { 4, LE_BIN, "Initial EIP" },
  { 4, LE_BIN, "Initial object SS number" },
  { 4, LE_BIN, "Initial ESP" },
  { 4, LE_BIN, "Memory page size" },
  { 4, LE_BIN, "Bytes on last page" },
  { 4, LE_BIN, "Fixup section size" },
  { 4, LE_BIN, "Fixup section checksum" },
  { 4, LE_BIN, "Loader section size" },
  { 4, LE_BIN, "Loader section checksum" },
  { 4, LE_BIN, "Offset of object table" },
  { 4, LE_BIN, "Object table entries" },
  { 4, LE_BIN, "Object page map offset" },
  { 4, LE_BIN, "Object iterate data map offset" },
  { 4, LE_BIN, "Resource table offset" },
  { 4, LE_BIN, "Resource table entries" },
  { 4, LE_BIN, "Resident names table offset" },
  { 4, LE_BIN, "Entry table offset" },
  { 4, LE_BIN, "Module directives table offset" },
  { 4, LE_BIN, "Module directives entries" },
  { 4, LE_BIN, "Fixup page table offset" },
  { 4, LE_BIN, "Fixup record table offset" },
  { 4, LE_BIN, "Imported modules name table offset" },
  { 4, LE_BIN, "Imported modules count" },
  { 4, LE_BIN, "Imported procedure name table offset" },
  { 4, LE_BIN, "Per-page checksum table offset" },
  { 4, LE_BIN, "Data pages offset from top of file" },
  { 4, LE_BIN, "Preload page count" },
  { 4, LE_BIN, "Non-resident names table offset from top of file" },
  { 4, LE_BIN, "Non-resident names table length" },
  { 4, LE_BIN, "Non-resident names table checksum" },
  { 4, LE_BIN, "Automatic data object" },
  { 4, LE_BIN, "Debug information offset" },
  { 4, LE_BIN, "Debug information length" },
  { 4, LE_BIN, "Preload instance pages number" },
  { 4, LE_BIN, "Demand instance pages number" },
  { 4, LE_BIN, "Extra heap allocation" },
  { 4, LE_BIN, "Unknown" },
  { 0 }
};

hdr_desc le_hot_tbl[] = {
  { 4, LE_BIN, "Virtual segment size in bytes" },
  { 4, LE_BIN, "Relocation base address" },
  { 4, LE_BIN, "Object flags" },
  { 4, LE_BIN, "Page map index" },
  { 4, LE_BIN, "Page map entries" },
  { 4, LE_STR, "Unknown" },
  { 0 }
};

void le_dump( char *b, int len, hdr_desc *tbl ) {
  int offs = 0;
 
  for ( int i = 0; tbl[i].size; i++ ) {
  	if ( offs >= len ) {
		puts( "premature end of table" );
		return;
	}
  	printf( "%04x\t", offs );
  	if ( tbl[i].type == LE_STR ) {
		putchar( '\'' );
		for ( int j = 0; j < tbl[i].size; j++ )
			putchar( b[offs + j] );
		printf( "'\t" );
	}
	else {
		if ( tbl[i].size == 1 )
			printf( "%02x\t", *(unsigned char*)&b[offs] );
		else if ( tbl[i].size == 2 )
			printf( "%04x\t", *(unsigned short*)&b[offs] );
		else if ( tbl[i].size == 4 )
			printf( "%08x", *(unsigned int*)&b[offs] );
		else	printf( "bin" );
	}
	printf( "\t%s\n", tbl[i].desc );
  	offs += tbl[i].size;
  }
}

int le_ent_points_dump( int fd, int offs ) {
  char *buf = 0;
  int ent_count, bnd_flags, bnd_size;

  int rc = lseek( fd, offs, SEEK_SET );
  if ( rc == -1) {
  	printf( "lseek set to %d\n", offs );
	perror( "lseek" );
	return 1;
  }

  int it_offs = 0;
  for (;;) {
	buf = (char*) realloc( buf, 2 );
 // reading header of entries bundle
	rc = read( fd, buf, 2 );
	if ( rc == -1 ) {
  		perror( "read" );
		return 1;
	}
	if ( rc < 2 ) {
  		printf( "read %d bytes of %s\n", rc, sizeof(buf) );
		return 1;
	}

	ent_count = *(unsigned char*)&buf[0];
	bnd_flags = *(unsigned char*)&buf[1];
	printf( "%04x\t%02x\t\tBundle items\n", it_offs, ent_count );
	printf( "%04x\t%02x\t\tBundle flags\n", it_offs + 1, bnd_flags );
	if ( !ent_count ) break;

	if ( bnd_flags & 2 ) bnd_size = ent_count * 4 + 2;
	else bnd_size = ent_count * 2 + 2;
	
	it_offs += 2;
// reading bundle entries
	buf = (char*) realloc( buf, bnd_size );
	rc = read( fd, buf, bnd_size );
	if ( rc == -1 ) {
  		perror( "read" );
		return 1;
	}
	if ( rc < bnd_size ) {
  		printf( "read %d bytes of %s\n", rc, bnd_size );
		return 1;
	}
	printf( "%04x\t%04x\t\tObject index\n",
		it_offs, *(unsigned short*)buf );
	it_offs += 2;
	int bidx = 2;
	for ( int i = 0; i < ent_count; i++ ) {
		printf( "%04x\t%02x\t\tEntry flags\n",
			it_offs++, *(unsigned char*)&buf[bidx] );
		if ( bnd_flags & 2 ) {
			printf( "%04x\t%08x\tSegment offset\n",
				it_offs, *(unsigned int*)&buf[bidx + 1] );
			it_offs += 4;
			bidx += 4;
		}			
		else {
			printf( "%04x\t%04x\t\tSegment offset\n",
				it_offs, *(unsigned short*)&buf[bidx + 1] );
			it_offs += 2;
			bidx += 2;
		}
	}
  }

  return 0;
}

int process_fixup_record( int reltype, char* &frptr, char *fdata ) {
  char *ptrsave;
  int ordval, addval;

  switch (reltype & 3) {
  case 0:
	printf( "%08x\t%02x\t\tTarget segment number\n",
		frptr - fdata, *(unsigned char*)frptr );
	frptr++;
	if ( reltype & 0x10 ) {
		printf( "%08x\t%08x\tTarget offset\n",
			frptr - fdata, *(unsigned int*)frptr );
		frptr += 4;
	}
	else {
		printf( "%08x\t%04x\t\tTarget offset\n",
			frptr - fdata, *(unsigned short*)frptr );
		frptr += 2;
	}
	break;
  case 1:
	printf( "%08x\t%02x\t\tImported module-name index\n",
		frptr - fdata, *(unsigned char*)frptr );
	frptr++;
	ptrsave = frptr;
	if ( reltype & 0x80 ) {
		ordval = *(unsigned char*)frptr;
		frptr++;
	}
	else {
		ordval = *(unsigned short*)frptr;
		frptr += 2;
	}
	printf( "%08x\t%04x\t\tOrdinal Value\n",
		ptrsave - fdata, ordval );

	if ( reltype & 4 ) {
		ptrsave = frptr;
		if ( reltype & 0x10 ) {
			addval = *(unsigned int*)frptr;
			frptr += 6;
		}
		else {
			addval = *(unsigned short*)frptr;
			frptr += 2;
		}
		printf( "%08x\t%08x\t\tAbs Add Value\n",
			ptrsave - fdata, addval );
	}
	break;
  case 2:
	printf( "%08x\t%02x\t\tImported module-name index\n",
		frptr - fdata, *(unsigned char*)frptr );
	printf( "%08x\t%02x\t\tOffset of name in Imported_Proc_Name_Table\n",
		frptr - fdata + 2, *(unsigned short*)&frptr[2] );
	frptr += 4;
	break;
  case 3:
	puts( "unsupported OSFIXUP" );
	return 0;
  }
  return 1;
}

void le_fixup_dump( int fd, int hdr_offs, char *hdr ) {
  int secsize, mapoffs, recoffs, pgcount, i, j;
  char *fdata, *frstart, *frptr;
  int *mptr;
  
  pgcount = *(unsigned int*)&hdr[0x14];
  secsize = *(unsigned int*)&hdr[0x30];
  mapoffs = *(unsigned int*)&hdr[0x68];
  recoffs = *(unsigned int*)&hdr[0x6c];

  fdata = (char*) malloc( secsize );
  lseek( fd, hdr_offs + mapoffs, SEEK_SET );
  read( fd, fdata, (secsize + 1) * 4);
  
  mptr = (int*) fdata;
  frstart = fdata + (recoffs - mapoffs);
  for ( i = 0; i < pgcount; i++ ) {
  	int reltype;

  	printf( "Page %d Fixups\n", i );
	frptr = (char*) (fdata + recoffs - mapoffs + mptr[i]);
	while ( frptr < fdata + recoffs - mapoffs + mptr[i + 1] ) {
  		if ( *frptr != 7 && *frptr != 0x27 &&
		     *frptr != 8 && *frptr != 0x28 )
		  	puts( "LOOK HERE >>>>" );
		printf( "%08x\t%01x\t\tRelocation address type\n",
			frptr - fdata, *(unsigned char*)frptr );
		reltype = *(unsigned char*)&frptr[1];
		printf( "%08x\t%01x\t\tRelocation type\n",
			frptr - fdata + 1, reltype );
		reltype = *(unsigned char*)&frptr[1];
		if ( !((*(unsigned char*)frptr) & 0x20) ) {
			printf( "%08x\t%04x\t\tOffset within page %d\n",
				frptr - fdata + 2, *(unsigned short*)&frptr[2], i );
			frptr += 4;
			if ( !process_fixup_record( reltype, frptr, fdata ) )
				goto next_page;
		}
		else {
			int offsnum, ordval, addval;
			char *ptrsave;
			
			offsnum = *(unsigned char*)&frptr[2];
			printf( "%08x\t%02x\t\tOffset counter\n",
				frptr - fdata + 2, offsnum );
			frptr += 3;
			if ( !process_fixup_record( reltype, frptr, fdata ) )
				goto next_page;
			for ( j = 0; j < offsnum; j++ ) {
				printf( "%08x\t%04x\t\tOffset within page %d\n",
					frptr - fdata, *(unsigned short*)frptr, i );
				frptr += 2;
			}
		}
	}
next_page:;
  }
}

int main( int argc, char **argv ) {
  char buf[0xb0], hdr[0xb0];
  int header_offs;

  if ( argc < 2 ) {
  	printf( "LE-Dump 1.0 by VR\n" );
  	printf( "usage: %s filename.vxd\n", basename( argv[0] ) );
	return 1;
  }
  int fd = open( argv[1], O_RDONLY );
  if ( fd == -1) {
	perror( "open" );
	return 1;
  }
 
 // reading pointer to LE header information block
  int rc = lseek( fd, 0x3c, SEEK_SET );
  if ( rc == -1) {
  	printf( "lseek set to %d\n", 0x3c );
	perror( "lseek" );
	return 1;
  }

  rc = read( fd, &header_offs, 4 );
  if ( rc == -1 ) {
  	perror( "read" );
	return 1;
  }
  if ( rc < 4 ) {
  	printf( "read %d bytes of %s\n", rc, 4 );
	return 1;
  }
  
  rc = lseek( fd, header_offs, SEEK_SET );
  if ( rc == -1) {
  	printf( "lseek set to %d\n", *(int*)buf );
	perror( "lseek" );
	return 1;
  }
 
 // reading LE header information block
  rc = read( fd, hdr, sizeof(buf) );
  if ( rc == -1 ) {
  	perror( "read" );
	return 1;
  }
  if ( rc < sizeof(hdr) ) {
  	printf( "read %d bytes of %s\n", rc, sizeof(hdr) );
	return 1;
  }

  puts( "LE header information block" );
  le_dump( hdr, 0xb0, le_hib_tbl );
  putchar( '\n' );

 // reading LE header object table
  unsigned int hot_offs = header_offs + *(unsigned int*)&hdr[0x40],
	hot_count = *(unsigned int*)&hdr[0x44],
	ent_offs = header_offs + *(unsigned int*)&hdr[0x5c],
	modimp_offs = header_offs + *(unsigned int*)&hdr[0x70],
	modimp_count = header_offs + *(unsigned int*)&hdr[0x74];

  rc = lseek( fd, hot_offs, SEEK_SET );
  if ( rc == -1) {
  	printf( "lseek set to %d\n", hot_offs );
	perror( "lseek" );
	return 1;
  }

  for ( int i = 0; i < hot_count; i++ ) {
	rc = read( fd, buf, 0x18 );
	if ( rc == -1 ) {
	  	perror( "read" );
		return 1;
	}
	if ( rc < 0x18 ) {
	  	printf( "read %d bytes of %s\n", rc, 0x18 );
		return 1;
	}
 
	printf( "LE header object #%d table\n", i + 1 );
	le_dump( buf, 0x18, le_hot_tbl );
  }
  putchar( '\n' );
  
  puts( "Entry points map" );
  le_ent_points_dump( fd, ent_offs );
  putchar( '\n' );

  puts( "Fixups section" );
  le_fixup_dump( fd, header_offs, hdr );
}
