#include <stdio.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include "sg.h"
#include "winvfs.h"

void scsiport_cd_entry( IOP *iop );

int sg_fd;

void scsiport_init( DCB *dcb ) {
  sg_fd = open( "/dev/sg0", O_RDWR );
  if ( sg_fd == -1 ) {
  	perror( "open sg0" );
	return;
  }

  DCB_cd_entry *ent = new DCB_cd_entry;
  ent->DCB_cd_io_address = scsiport_cd_entry;
  ent->DCB_cd_flags = 0;
  ent->DCB_cd_ddb = 0;
  ent->DCB_cd_next = dcb->DCB_cmn.DCB_ptr_cd;
  ent->DCB_cd_expan_off = 0;
  ent->DCB_cd_layer_flags = 0;
  
  dcb->DCB_cmn.DCB_ptr_cd = ent;
}

void scsiport_cd_entry( IOP *iop ) {
  int i, rc;
  IOR *ior = &iop->IOP_ior;
  SCSI_REQUEST_BLOCK *srb = (SCSI_REQUEST_BLOCK*) iop->IOP_srb;

  if ( ior->IOR_func != 0x0f ) {	// scsi_pass_through
  	iop->IOP_calldown_ptr = iop->IOP_calldown_ptr->DCB_cd_next;
  	iop->IOP_calldown_ptr->DCB_cd_io_address( iop );
	return;
  }

  sg_io_hdr sg;
  memset( &sg, 0, sizeof(sg) );
  sg.interface_id = 'S';
  if ( srb->SrbFlags & 0x40 )
	sg.dxfer_direction = SG_DXFER_FROM_DEV;
  else if ( srb->SrbFlags & 0x80 )
  	sg.dxfer_direction = SG_DXFER_TO_DEV;
  else sg.dxfer_direction = SG_DXFER_NONE;
  sg.cmd_len = srb->CdbLength;
  sg.mx_sb_len = srb->SenseInfoBufferLength;
  sg.dxfer_len = srb->DataTransferLength;
  sg.dxferp = srb->DataBuffer;
  sg.cmdp = srb->Cdb;
  sg.sbp = (unsigned char*) srb->SenseInfoBuffer;
  sg.timeout = 0x1c00; // srb->TimeOutValue;

  printf( "SCSIPORT srb=0x%08x, srb->DataBuffer=0x%08x srb->SenseBuffer=0x%08x\n",
  	srb, srb->DataBuffer, srb->SenseInfoBuffer );
  printf( "SCSIPORT srb->SrbFlags=0x%08x\n",
  	srb->SrbFlags );
  printf( "scsi cmd:" );
  for ( i = 0; i < sg.cmd_len; i++ ) printf( " %02X", sg.cmdp[i] );
  putchar( '\n' );

  rc = ioctl( sg_fd, SG_IO, &sg );

  printf( "scsi rc=%d, status=%d\n", rc, sg.status );
  if ( sg.sb_len_wr ) {
  	printf( "sense:" );
	for ( i = 0; i < sg.sb_len_wr; i++ ) printf( " %02X", sg.sbp[i] );
	putchar( '\n' );
  }

  if ( sg.dxfer_len ) {
  	printf( "data1 (%db):", sg.dxfer_len );
	for ( i = 0; i < sg.dxfer_len && i < 20; i++ )
		printf( " %02X", ((unsigned char*)sg.dxferp)[i] );
	putchar( '\n' );
  }

  if ( sg.status == 0 ) srb->SrbStatus = 1;	// SRB_STATUS_SUCCESS
  else {
  	srb->SrbStatus = 4;			// SRB_STATUS_ERROR
	printf( "SCSIPORT status=0x%x\n", sg.status );
	sleep(3);
  }
  srb->ScsiStatus = sg.status;
  srb->SenseInfoBufferLength = sg.sb_len_wr;

  iop->IOP_ior.IOR_status = 0;			// IORS_SUCCESS
  iop->IOP_ior.IOR_callback( iop );
}
