#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fuse.h>
#include <errno.h>
#include "vxdload.h"
#include "winvfs.h"

extern volfunc *vfs_vfunc;
extern void *vfs_rh;
extern vxd_internal *vxd;
extern int fuse_fd;

void print_parsed_path( ParsedPath *pp ) {
  int len, sl, i;
  unsigned short *p;

  printf( "total length=%d\n", pp->pp_totalLength );
  printf( "prefix length=%d\n", pp->pp_prefixLength );
  p = (unsigned short*) pp->pp_elements;
  for ( len = 4; len < pp->pp_totalLength; len += sl ) {
  	sl = *p++;
  	printf( "segment: %d,%d,'", (int)p - (int)pp, sl );
	for ( i = 0; i < (sl - 2) / 2; i++ ) putchar( *p++ );
	printf( "'\n" );
  }
}

ParsedPath *parse_path( char *path ) {
  ParsedPath *pp;
  int bp, len, sl, i;
  unsigned short *buf;
  unsigned char *p, *s;
  
  len = 2;
  buf = (unsigned short*) malloc( len * sizeof(short) );
  pp = (ParsedPath*) buf;
  pp->pp_totalLength = 4;
  pp->pp_prefixLength = 4;
  bp = 2;	// pp->ppElements
  p = (unsigned char*) path;
  for (;;) {
	s = (unsigned char*) strchr( (char*) p, '/' );
	if ( s ) sl = s - p; else sl = strlen( (char*) p );
	buf = (unsigned short*) realloc( buf, (len + sl + 1) * sizeof(short) );
	pp = (ParsedPath*) buf;
	pp->pp_totalLength += (sl+1) * 2;
	len += sl + 1;
	buf[bp++] = (sl+1) * 2;
	for ( i = 0; i < sl; i++ ) buf[bp++] = *p++;
	if ( !s ) break;
	p = s + 1;
	pp->pp_prefixLength += (sl+1) * 2;
  }

  return pp;
}

void parsed_path_append( ParsedPath* &pp, char *entry ) {
  unsigned short *p;
  int i, offs, len = strlen( entry );
 
  p = (unsigned short*) ((int)pp + pp->pp_prefixLength);
  if ( *p == 2 )
  	pp->pp_totalLength = pp->pp_prefixLength + len * 2 + 2;
  else {
  	pp->pp_totalLength += len * 2 + 2;
	pp->pp_prefixLength += *p;
	p = (unsigned short*)((char*)p + *p);
  }
  offs = (int)p - (int)pp;
  pp = (ParsedPath*) realloc( pp, pp->pp_totalLength );
  p = (unsigned short*) ((int)pp + offs);
  *p++ = (len+1) * 2;
  for ( i = 0; i < len; i++ ) *p++ = entry[i];
}

char *get_cstr_from_uni( unsigned short *ustr ) {
  int len;
  unsigned short *us;
  char *str, *p;

  for ( us = ustr, len = 0; *us; us++ ) len++;
  str = (char*) malloc( len + 1 );
  for ( us = ustr, p = str; *us; ) *p++ = *us++;
  *p++ = 0;
  return str;
}

int __winvfs_findopen( ioreq *ir, const char *path, char *wildcard ) {
  int rc, (*findOpen)(ioreq*);
  
  findOpen = (int (*)(ioreq*)) vfs_vfunc->vfn_func[13]; // VFN_FINDOPEN
  printf( "VFN_FINDOPEN at 0x%08x, ir_rh=0x%08x, path=%s\n",
        (int) findOpen - (int) vxd->sectbl[0], vfs_rh, path );
  memset( ir, 0, sizeof(ir) );
  ir->ir_rh = vfs_rh;
  ir->ir_pid = 1;
  ir->ir_attr = 0x50000010; // FILE_FLAG_WILDCARDS, FILE_FLAG_HAS_STAR,
                             // FILE_FLAG_LONG_PATH
  if ( wildcard ) ir->ir_attr |= 0x80000000; 
  ir->ir_data = (win32_find_data*) malloc( sizeof(win32_find_data) );
  ir->ir_ppath = parse_path( (char*) path + 1 );
  if ( wildcard ) parsed_path_append( ir->ir_ppath, wildcard );
  
  char *sp;
  unsigned short *p, *fname, *pname;
  fname = (unsigned short*) malloc( 2 * sizeof(short) );
  p = fname;
  *p++ = '*'; *p = 0;
  pname = (unsigned short*) malloc( strlen( path ) * sizeof(short) );
  for ( sp = (char*) path + 1, p = pname; *sp; ) *p++ = *sp++;
  *p = 0;
  ir->ir_uFName = fname;
  ir->ir_upath = pname;
 
  ir->ir_hfunc = (hndlfunc*) malloc( sizeof(hndlfunc) );
  memset( ir->ir_hfunc, 0, sizeof(hndlfunc) );

  rc = findOpen( ir );
  puts( "VFN_FINDOPEN done" );

  return rc;
}

int winvfs_getdir( const char *path, fuse_dirh_t h, fuse_dirfil_t filler ) {
  int rc, (*findNext)(ioreq*), (*findClose)(ioreq*);
  char *rname;
  ioreq ir;

  rc = __winvfs_findopen( &ir, path, "*" );
  if ( rc ) return -ENOENT;

  filler( h, ".", 0 );
  if ( !strcmp( path, "/" ) ) filler( h, "..", 0 );
  rname = get_cstr_from_uni( ((win32_find_data*)ir.ir_data)->cFileName );
  if ( strlen(rname) ) filler( h, rname, 0 );
  free( rname );

  findNext = (int (*)(ioreq*)) ir.ir_hfunc->hf_read;
  findClose = (int (*)(ioreq*)) ir.ir_hfunc->hf_misc->hm_func[1];
  ir.ir_options = 0;
  printf( "FindNext at 0x%08x, ir_fh=0x%08x\n",
        (int) findNext - (int) vxd->sectbl[0], ir.ir_fh );
  for (;;) {
  	rc = findNext( &ir );
	if ( rc ) break;
	rname = get_cstr_from_uni( ((win32_find_data*)ir.ir_data)->cFileName );
	if ( strlen(rname) ) filler( h, rname, 0 );
	free( rname );
  }

  printf( "FindClose at 0x%08x, ir_fh=0x%08x\n",
        (int) findClose - (int) vxd->sectbl[0], ir.ir_fh );
  findClose( &ir );

  return 0;
}

int __winvfs_open( ioreq *ir, char *path, int flags ) {
  int rc, (*reqFunc)(ioreq*);

  reqFunc = (int (*)(ioreq*)) vfs_vfunc->vfn_func[5]; // VFN_OPEN
  printf( "VFN_OPEN at 0x%08x, ir_rh=0x%08x\n",
        (int) reqFunc - (int) vxd->sectbl[0], vfs_rh );
  memset( ir, 0, sizeof(ir) );
  ir->ir_rh = vfs_rh;
  ir->ir_pid = 1;
  ir->ir_options = 1;   // ACTION_OPENEXISTING
  ir->ir_ppath = parse_path( path + 1 );
  print_parsed_path( ir->ir_ppath );
  
  char *sp;
  unsigned short *p, *fname, *pname;
  fname = (unsigned short*) malloc( 2 * sizeof(short) );
  p = fname;
  *p++ = '*'; *p = 0;
  pname = (unsigned short*) malloc( strlen( path ) * sizeof(short) );
  for ( sp = (char*) path + 1, p = pname; *sp; ) *p++ = *sp++;
  *p = 0;
  ir->ir_uFName = fname;
  ir->ir_upath = pname;
  
  ir->ir_hfunc = (hndlfunc*) malloc( sizeof(hndlfunc) );
  memset( ir->ir_hfunc, 0, sizeof(hndlfunc) );

  rc = reqFunc( ir );
  printf( "VFN_OPEN rc=0x%x ir_error=0x%x\n", rc, ir->ir_error );
  return rc;
}

int __winvfs_close( ioreq *ir ) {
  int (*closeFile)(ioreq*);

  ir->ir_flags = 2; // CLOSE_FINAL
  closeFile = (int (*)(ioreq*)) ir->ir_hfunc->hf_misc->hm_func[1];
  printf( "WINVFS_close at 0x%08x ir_fh=0x%08x\n",
	(int) closeFile - (int) vxd->sectbl[0], ir->ir_fh );
  closeFile( ir );

  return 0;
}

int winvfs_getattr( const char *path, struct stat *stbuf ) {
  int (*findClose)(ioreq*);
  ioreq ir;
  
  memset( stbuf, 0, sizeof(struct stat) );
  stbuf->st_nlink = 2;
  stbuf->st_mode = S_IFDIR | 0555;
  stbuf->st_size = 1024;
  if ( !strcmp( path, "/" ) ) return 0;

  if ( __winvfs_findopen( &ir, path, 0 ) ) return -ENOENT;
  findClose = (int (*)(ioreq*)) ir.ir_hfunc->hf_misc->hm_func[1];
  findClose( &ir );

  int attr = ((win32_find_data*)ir.ir_data)->dwFileAttributes;
  if ( !(attr & 0x10) ) {
  	stbuf->st_nlink = 1;
	stbuf->st_mode = S_IFREG | 0444;
	stbuf->st_size = ((win32_find_data*)ir.ir_data)->nFileSizeLow;
  }

  return 0;
}

int winvfs_open( const char *path, int flags ) {
  ioreq ir;
  
  if ( __winvfs_open( &ir, (char*) path, flags ) ) return -ENOENT;
  __winvfs_close( &ir );
  
  return 0;
}

int winvfs_read( const char *path, char *buf, size_t size, off_t offset ) {
  ioreq ir;
  int pos, rc, (*readFile)(ioreq*);
  
  if ( __winvfs_open( &ir, (char*) path, 0 ) ) return -ENOENT;
  readFile = (int (*)(ioreq*)) ir.ir_hfunc->hf_read;

  printf( "FS_ReadFile path=%s, offset=%d, size=%d\n", path, offset, size );
  ir.ir_pos = offset;
  ir.ir_data = buf;
  ir.ir_length = size;
  rc = readFile( &ir );
  if ( rc ) rc = 0; else rc = ir.ir_length;
  __winvfs_close( &ir );

  return rc;
}

void winvfs_fuse() {
  fuse_operations winvfs_ops;
  
  memset( &winvfs_ops, 0, sizeof(winvfs_ops) );
  winvfs_ops.getattr = winvfs_getattr;
  winvfs_ops.getdir = winvfs_getdir;
  winvfs_ops.open = winvfs_open;
  winvfs_ops.read = winvfs_read;

  fuse_loop( fuse_new( fuse_fd, 0, &winvfs_ops ) );
}
