#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include "vxdload.h"

void apply_fixup( char **sectbl, char *srcptr,
	unsigned char addrtype, unsigned char reltype, char *frptr ) {
  char *target;

  if ( (reltype & 3) == 0 ) {
	if ( reltype & 0x10 )
		target = sectbl[(*(unsigned char*)frptr) - 1] + *(unsigned int*)&frptr[1];
	else
		target = sectbl[(*(unsigned char*)frptr) - 1] + *(unsigned short*)&frptr[1];
	if ( (addrtype & 15) == 7 )
		*(char**)srcptr = target;
	else if ( (addrtype & 15) == 8 )
		*(int*)srcptr = target - srcptr - 4;
  }
  else
  	printf( "reltype=%d for fixups not implemented\n" );
}

char *fixup_skip_target( char *frptr, unsigned char reltype ) {
  if ( (reltype & 3) == 0 ) {
  	if ( reltype & 0x10 ) return frptr + 5; else return frptr + 3;
  }
  else return frptr;
}

int page_fixup( char *hdr, char **sectbl,
	 char *fxdata, char *page, int pgnum ) {
  int i, pgoffs;
  unsigned int *mptr, mapoffs, recoffs, offsnum;
  unsigned char addrtype, reltype;
  char *frptr;

  mapoffs = *(unsigned int*)&hdr[0x68];
  recoffs = *(unsigned int*)&hdr[0x6c];
  
  mptr = (unsigned int*) fxdata;
  frptr = fxdata + recoffs - mapoffs + mptr[pgnum];
  while ( frptr < fxdata + recoffs - mapoffs + mptr[pgnum + 1] ) {
	addrtype = *(unsigned char*)frptr;	
	reltype = *(unsigned char*)&frptr[1];
	if ( addrtype != 7 && addrtype != 0x27 &&
		 addrtype != 8 && addrtype != 0x28 ) {
		printf( "\nsomething went wrong on page %d fixup_offs=%08x addrtype=%02x\n",
			pgnum, frptr - fxdata, addrtype );
		return -1;
	}
	frptr += 2;
	if ( !(addrtype & 0x20) ) {
		pgoffs = *(short*)frptr;
		apply_fixup( sectbl, page + pgoffs, addrtype, reltype, frptr + 2 );
		frptr = fixup_skip_target( frptr + 2, reltype );
	}
	else {
		offsnum = *(unsigned char*)frptr;
		frptr++;
		for ( i = 0; i < offsnum; i++ ) {
			pgoffs = *(short*)(fixup_skip_target(frptr, reltype) + i * 2);
			apply_fixup( sectbl, page + pgoffs, addrtype, reltype, frptr );
		}
		frptr = fixup_skip_target( frptr, reltype ) + offsnum * 2;
	}
  }
  return 0;
}

int load_section( int fd, int hdr_offs, char *hdr,
	char **sechdrl, char **sectbl, int secnum, char *fxdata ) {
  int idxptr, startpg, pgnum, pgsize, dataoffs, j, page;
  char *pgtbl, *sechdr;

  sechdr = sechdrl[secnum];
  startpg = *(unsigned long*)&sechdr[0x0c] - 1;
  idxptr = hdr_offs + *(unsigned long*)&hdr[0x48] + startpg * 4;
  pgnum = *(unsigned long*)&sechdr[0x10];
  lseek( fd, idxptr, SEEK_SET );
  pgtbl = (char*) malloc( pgnum * 4 );
  read( fd, pgtbl, pgnum * 4 );
  
  dataoffs = *(unsigned long*)&hdr[0x80];
  pgsize = *(unsigned long*)&hdr[0x28];
  for ( j = 0; j < pgnum; j++ ) {
  	page = *(unsigned short*)&pgtbl[j * 4] +
		*(unsigned char*)&pgtbl[j * 4 + 2] - 1;
	lseek( fd, dataoffs + page * pgsize, SEEK_SET );
	read( fd, sectbl[secnum] + pgsize * j, pgsize );
	if ( page_fixup( hdr, sectbl, fxdata, sectbl[secnum] + pgsize * j, startpg + j ) == -1 ) {
		free( pgtbl );
		return 0;
	}
  }
  free( pgtbl );
  return 1;
}

#pragma pack(1)

struct vdd_tbl {
  int DDB_Next;
  short DDB_SDK_Version;
  short DDB_Req_Device_Number;
  char DDB_Dev_Major_Version;
  char DDB_Dev_Minor_Version;
  short DDB_Flags;
  char DDB_Name[8];
  int DDB_Init_Order;
  int DDB_Control_Proc;
  int DDB_V86_API_Proc;
  int DDB_PM_API_Proc;
  int DDB_V86_API_CSIP;
  int DDB_PM_API_CSIP;
  int DDB_Reference_Data;
  int *DDB_Service_Table_Ptr;
  int DDB_Service_Table_Size;
};

vdd_tbl *find_vdd( int fd, char *hdr, int hdr_offs, char **sectbl ) {
  char *buf, *entry;
  int ent_count, flags, bnd_size;

  buf = (char*) malloc( 2 );
  lseek( fd, hdr_offs + *(unsigned int*)&hdr[0x5c], SEEK_SET );
  read( fd, buf, 2 );
  ent_count = *(unsigned char*)buf;
  flags = *(unsigned char*)&buf[1];

  if ( flags & 2 ) bnd_size = ent_count * 4 + 3;
  else bnd_size = ent_count * 2 + 3;
  buf = (char*) realloc( buf, bnd_size );
  read( fd, buf, bnd_size );
  if ( flags & 2 )
	entry = sectbl[*(unsigned short*)buf - 1] + *(unsigned int*)&buf[3];
  else
	entry = sectbl[*(unsigned short*)buf - 1] + *(unsigned short*)&buf[3];
  free( buf );
  return (vdd_tbl*) entry;
}

void load_servtbl( vdd_tbl *vdd, vxd_internal *vxd ) {
  vxd->servnum = vdd->DDB_Service_Table_Size;
  vxd->serv = (int*) malloc( vxd->servnum * sizeof(int) );
  for ( int i = 0; i < vxd->servnum; i++ )
	vxd->serv[i] = vdd->DDB_Service_Table_Ptr[i];
}

vxd_internal *load_vxd( char *name ) {
  int fd, fd2, rc, i, j;
  char *p;
 
  printf( "loading %s...\n", name ); 
  fd = open( name, O_RDONLY );
  if ( fd == -1 ) {
  	perror( "load_vxd.cc: open" );
	return 0;
  }

  int hdr_offs, pgsize, pgnum, fxsize, mapoffs;
  char hdr[0xb0], *fxdata;
  vxd_internal *vxd;

  vxd = new vxd_internal;
  memset( vxd, 0, sizeof(vxd_internal) );

// main header  
  lseek( fd, 0x3c, SEEK_SET );
  read( fd, &hdr_offs, 4 );
  lseek( fd, hdr_offs, SEEK_SET );
  read( fd, hdr, sizeof(hdr) );

// fixups
  fxsize = *(unsigned int*)&hdr[0x30];
  mapoffs = *(unsigned int*)&hdr[0x68];
  fxdata = (char*) malloc( fxsize );
  lseek( fd, hdr_offs + mapoffs, SEEK_SET );
  read( fd, fxdata, fxsize );
 
// sections
  vxd->secnum = *(unsigned long*)&hdr[0x44];
  vxd->sechdr = (char**) malloc( vxd->secnum * sizeof(char*) );
  vxd->sectbl = (char**) malloc( vxd->secnum * sizeof(char*) );
  vxd->secsz = (int*) malloc( vxd->secnum * sizeof(int) );
  pgsize = *(unsigned long*)&hdr[0x28];
  for ( i = 0; i < vxd->secnum; i++ ) {
	vxd->sechdr[i] = (char*) malloc( 0x18 );
  	lseek( fd, hdr_offs + *(unsigned long*)&hdr[0x40] +
		i * 0x18, SEEK_SET );
  	read( fd, vxd->sechdr[i], 0x18 );
  	pgnum = *(unsigned long*)&vxd->sechdr[i][0x10];
	vxd->sectbl[i] = (char*) malloc( pgsize * pgnum );
	vxd->secsz[i] = pgsize * pgnum;
  }

  for ( i = 0; i < vxd->secnum; i++ ) {
	printf( "fixing up %-4.4s at 0x%08x... ",
		&vxd->sechdr[i][0x14], vxd->sectbl[i] );
	fflush( stdout );
	rc = load_section( fd, hdr_offs, hdr, vxd->sechdr, vxd->sectbl, i, fxdata );
	if ( !rc ) {
		puts( "failed" );
		return 0;
	}
	puts( "ok" );
  }

  vdd_tbl *vdd;
  vdd = find_vdd( fd, hdr, hdr_offs, vxd->sectbl );
  close( fd );
  vxd->ctl_disp = (void (*)(int)) vdd->DDB_Control_Proc;
  vxd->devnum = vdd->DDB_Req_Device_Number;
  load_servtbl( vdd, vxd );

  memcpy( vxd->name, vdd->DDB_Name, 8 );
  p = &vxd->name[7];
  while ( *p == ' ' && p >= vxd->name ) p--;
  p[1] = 0;
  printf( "control_proc of %s points to 0x%08x\n", vxd->name, vxd->ctl_disp );
  for ( i = 0; i < vxd->servnum; i++ )
	printf( "service %04X.%04X is at 0x%08x\n",
		vxd->devnum, i, vxd->serv[i] - (int) vxd->sectbl[0] );
  putchar( '\n' );

  return vxd;
}
