// cleanup Device_Service_Hooks:
//   SIGSEGV handler should be able to access hooked services list, i.e.
//   Hook_Device_Service should be executed in control process context,
//   or overall design should be not process, but threads based. 

#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif
#include <stdio.h>
#include <errno.h>
#include <malloc.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <unistd.h>
#include <libgen.h>
#include <signal.h>
#include <stdarg.h>
#include <ctype.h>
#include <iconv.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/ptrace.h>
#include <linux/user.h>
#include <list>
#include <vector>

#include "vxdload.h"
#include "winvfs.h"

void scsiport_init( DCB *dcb );
ParsedPath *parse_path( char* );
void print_parsed_path( ParsedPath* );
void parsed_path_append( ParsedPath*&, char* );
void winvfs_fuse();

// registered callbacks, structs and services
vector <vxd_internal*> vxd_list;
vxd_internal *vxd;
volfunc *vfs_vfunc = 0;
void *vfs_rh = 0;

struct DCB_cdrom;
DCB_cdrom *cdr_dcb_ph, *cdr_dcb_lg;
struct VRP;
VRP *vrp;

// VMM subsystem
bool uni_verbose=0, mtx_verbose=0, list_verbose=0, heap_verbose=0;

int VMM_PagerRegister( int ppd ) {
  printf( "VMM.PagerRegister (ppd=0x%08x)\n", ppd );
  return 0;
}

int VMM_CallRestrictedEvent( int boost, int hVMorThread, int flags,
	int refData, int pCallback, int timeout, int pThunk ) {
  printf( "VMM.CallRestrictedEvent (boost=%d, hVMorThread=%d, flags=0x%08x, "
	  "refData=0x%08x, pCallback=0x%08x, timeout=%d, pThunk=0x%08x)\n",
	boost, hVMorThread, flags, refData, pCallback, timeout, pThunk );
  return 0;
}

enum HKEY_enum {
  HKEY_CLASSES_ROOT,
  HKEY_CURRENT_USER,
  HKEY_LOCAL_MACHINE,
  HKEY_USERS,
  HKEY_CURRENT_CONFIG,
  HKEY_DYN_DATA
};

char *HKEY_str[] = {
  "HKEY_CLASSES_ROOT",
  "HKEY_CURRENT_USER",
  "HKEY_LOCAL_MACHINE",
  "HKEY_USERS",
  "HKEY_CURRENT_CONFIG",
  "HKEY_DYN_DATA"
};

struct regkey_internal {
  HKEY_enum root;
  char *key;
  int handle;
};

int regkeys_count = 4;
regkey_internal registry[] = {
  { HKEY_LOCAL_MACHINE, "Software\\Adaptec\\CDR4VSD", 10 },
  { HKEY_LOCAL_MACHINE, "Software\\Adaptec\\DirectCD", 11 },
  { HKEY_LOCAL_MACHINE, "Software\\Classes\\AudioCD\\shell", 12 },
//  { HKEY_LOCAL_MACHINE, "SYSTEM\\CurrentControlSet\\Services\\Cdrpwd\\Device0\\Parameters", 13 }, no key
  { HKEY_CURRENT_USER, "Software\\Microsoft\\Windows\\CurrentVersion\\Policies\\Explorer", 15 }
};

// values
// HKEY_LOCAL_MACHINE\SOFTWARE\Adaptec\DirectCD FSDLazyWritePeriod (no value)
// HKEY_LOCAL_MACHINE\SOFTWARE\Adaptec\CDR4VSD Enabled REG_BIN 01 00 00 02

int VMM_RegOpenKey( HKEY_enum hkey, char *lpszSubKey, int *phkResult ) {
  printf( "VMM.RegOpenKey %s\\%s\n", HKEY_str[hkey], lpszSubKey );

  for ( int i = 0; i < regkeys_count; i++ ) {
	if ( (unsigned char) hkey == registry[i].root &&
	     !strcasecmp( lpszSubKey, registry[i].key ) ) {
		*phkResult = registry[i].handle;
		return 0;
	}
  }
 
  puts( "VMM.RegOpenKey key not found" );
  return 2; // ERROR_FILE_NOT_FOUND
}

int VMM_RegCreateKey( HKEY_enum hkey, char *lpszSubKey, HKEY_enum *phkResult ) {
  printf( "VMM.RegCreateKey %s\\%s\n", HKEY_str[hkey], lpszSubKey );
  
  return 1016; // ERROR_REGISTRY_IO_FAILED
}

int VMM_GetRegistryKey( int dwType, char *lpszDeviceName,
	int dwFlags, int *lphDevKey ) {
  printf( "VMM.GetRegistryKey type=0x%08x flags=0x%08x devname=%s\n",
  	dwType, dwFlags, lpszDeviceName );
  
  return 2; // ERROR_FILE_NOT_FOUND
}

int VMM_RegDeleteKey( HKEY_enum hkey, char *lpszSubKey ) {
  printf( "VMM.RegDeleteKey %s\\%s\n", HKEY_str[hkey], lpszSubKey );
  return 0; // SUCCESS
}

int VMM_RegQueryValueEx( HKEY_enum hkey, char *lpszValueName,
	int *lpdwReserved, int *lpdwType, char *lpbData, int *lpcbData ) {
  printf( "VMM.RegQueryValueEx hkey=%d valueName=%s buf=0x%08x len=%d\n",
  	hkey, lpszValueName, lpbData, *lpcbData );
  return 2; // ERROR_FILE_NOT_FOUND
}

int VMM_RegSetValueEx( HKEY_enum hkey, char *lpszValueName,
	int dwReserved, int fdwType, char *lpbDta, int cbData ) {
  char *val;

  if ( fdwType == 1 )		// REG_SZ
  	val = strdup( lpbDta );
  else if ( cbData ) {		// REG_BINARY
  	val = (char*) malloc( cbData * 3 );
	for ( int i = 0; i < cbData; i++ )
		sprintf( val + i * 3, "%02X ", ((unsigned char*)lpbDta)[i] );
	
  }
  else	val = strdup( "(null)" );
  
  printf( "VMM.RegSetValueEx hkey=%d type=0x%08x valueName=%s "
  		"value=%s len=%d\n",
  	hkey, fdwType, lpszValueName, val, cbData );
  free( val );
  return 1016; // ERROR_REGISTRY_IO_FAILED
}

struct fx_list {
  int nsize;
  list <char*> *lst;
};

list <fx_list*> allocated_lists;

list <char*> *__VMM_ListCreate( int flags, int nodesize ) {
  fx_list *flst = new fx_list;
  flst->lst = new list <char*>;
  flst->nsize = nodesize;
  allocated_lists.push_back( flst );
  if ( list_verbose )
	printf( "VMM.ListCreate lst=0x%08x, nodesize=%d\n", flst->lst, nodesize );
  return flst->lst;
}

void __VMM_ListDestroy( list <char*> *lst ) {
  list <char*>::iterator it;
  list <fx_list*>::iterator it2;

  if ( list_verbose ) printf( "VMM.ListDestroy lst=0x%08x\n", lst );
  for ( it = lst->begin(); it != lst->end(); ++it ) free( *it );
  for ( it2 = allocated_lists.begin(); it2 != allocated_lists.end(); ++it2 )
	if ( (*it2)->lst == lst ) {
		delete *it2;
		allocated_lists.erase( it2 );
		break;
	}
  delete lst;
}

char *__VMM_List_Get_First( list <char*> *lst ) {
  if ( lst->begin() == lst->end() ) {
  	if ( list_verbose )
		printf( "VMM.List_Get_First lst=0x%08x, empty\n", lst );
	return 0;
  }

  if ( list_verbose )
	printf( "VMM.List_Get_First lst=0x%08x, item=0x%08x\n", lst, *lst->begin() ); 
  return *lst->begin();
}

char *__VMM_List_Get_Next( list <char*> *lst, char *node ) {
  list <char*>::iterator it;

  for ( it = lst->begin(); it != lst->end(); ++it ) {
  	if ( *it == node ) {
		++it;
		if ( it == lst->end() ) {
			if ( list_verbose )
	  			printf( "VMM.List_Get_Next lst=0x%08x, item=0x%08x, no more items\n", lst, node );
			return 0;
		}
		if ( list_verbose )
	  		printf( "VMM.List_Get_Next lst=0x%08x, item=0x%08x, next=0x%08x\n", lst, node, *it );
		return *it;
	}
  }
  if ( list_verbose )
	printf( "VMM.List_Get_Next lst=0x%08x, item=0x%08x, item not found\n", lst, node );
  return 0;
}

char *__VMM_ListAllocate( list <char*> *lst ) {
  list <fx_list*>::iterator it;
  char *item;

  for ( it = allocated_lists.begin(); it != allocated_lists.end(); ++it )
	if ( (*it)->lst == lst ) {
		item = (char*) malloc( (*it)->nsize );
		if ( list_verbose )
			printf( "VMM.ListAllocate lst=0x%08x, item=0x%08x, size=%d\n", lst, item, (*it)->nsize );
		memset( item, 0, (*it)->nsize );
		return item;
	}

  if ( list_verbose )
	printf( "VMM.ListAllocate lst=0x%08x, no such list\n", lst );
  return 0;
}

void __VMM_ListDeallocate( list <char*> *lst, char *node ) {
  if ( list_verbose )
  	printf( "VMM.ListDeallocate lst=0x%08x, item=0x%08x\n", lst, node );
  free( node );
}

bool __VMM_List_Remove( list <char*> *lst, char *node ) {
  list <char*>::iterator it;

  for ( it = lst->begin(); it != lst->end(); ++it )
	if ( *it == node ) {
		lst->erase( it );
		if ( list_verbose )
			printf( "VMM.List_Remove lst=0x%08x, item=0x%08x, succeeded\n", lst, node );
		return true;
	}
  if ( list_verbose )
	printf( "VMM.List_Remove lst=0x%08x, item=0x%08x, failed\n", lst, node );
  return false;
}

void __VMM_List_Attach( list <char*> *lst, char *node ) {
  if ( list_verbose )
	printf( "VMM.List_Attach lst=0x%08x, item=0x%08x\n", lst, node );
  lst->insert( lst->begin(), node );
}

unsigned int VMM_HeapAllocate( unsigned int nbytes, unsigned int flags ) {
  void *mem;

  mem = malloc( nbytes );
  memset( mem, 0, nbytes );
  if ( heap_verbose )
	printf( "VMM.HeapAllocate mem=0x%08x, nbytes=%d\n", mem, nbytes );
  return (unsigned int) mem;
}

unsigned int VMM_HeapFree( unsigned int hAddress, unsigned int flags ) {
  if ( heap_verbose )
	printf( "VMM.HeapFree mem=0x%08x\n", hAddress );
  free( (void*) hAddress );
  return 0;
}

int __VMM_PageGetAllocInfo( int flags ) {
  int npages;

  npages = 4 * 1024 * 1024 / getpagesize();
  printf( "VMM.PageGetAllocInfo return=%d pages (%d bytes per page)\n",
	npages, getpagesize() );
  return npages;
}

int __VMM_PageAllocate( int nPages, int pType, int hVM, int AlignMask,
	int minPhys, int maxPhys, int *PhysAddr, int Flags ) {
  int addr;

  addr = (int) memalign( 4096 * (AlignMask + 1), getpagesize() * nPages );
  *PhysAddr = addr;

  printf( "VMM.PageAllocate 1: pages=%d, ptype=0x%x, hvm=0x%0x\n",
	nPages, pType, hVM );
  printf( "VMM.PageAllocate 2: alignmask=0x%x, minPhys=0x%08x, maxPhys=0x%08x\n",
	AlignMask, minPhys, maxPhys );
  printf( "VMM.PageAllocate 3: flags=0x%x addr=0x%08x\n", Flags, addr );
  return addr;
}

unsigned int VMM_LinPageLock( unsigned int page,
	unsigned int npages, unsigned int flags ) {
  printf( "VMM.LinPageLock page=0x%08x, npages=%d, flags=0x%08x\n",
	page * getpagesize(), npages, flags );
  return page * getpagesize();
}

int VMM_Debug_Printf_Service( char *fmt, va_list ap ) {
  return vprintf( fmt, ap );
}

int VMM_CreateMutex( long boost, unsigned long flags ) {
  if ( mtx_verbose )
	  printf( "VMM.CreateMutex boost=0x%x, flags=0x%x\n", boost, flags );
  return 1;
}

int VMM_DestroyMutex( int hMutex ) {
  if ( mtx_verbose ) puts( "VMM.DestroyMutex" );
  return 1;
}

void VMM_EnterMutex( int hMutex, int flags ) {
  if ( mtx_verbose ) printf( "VMM.EnterMutex flags=0x%x\n", flags );
}

void VMM_LeaveMutex( int hMutex ) {
  if ( mtx_verbose ) puts( "VMM.LeaveMutex" );
}

int VMM_Sprintf( char *buf, char *fmt, ... ) {
  va_list ap;
  int rc;

  va_start( ap, fmt );
  rc = vsprintf( buf, fmt, ap );
  va_end( ap );
  return rc;
}

struct hook_internal {
  int service;
  void (*hook)();
};

list <hook_internal*> hooked_services;

int __VMM_Hook_Device_Service( int service, int hook0 ) {
  void (*hook)() = (void (*)()) hook0;
  printf( "VMM.Hook_Device_Service service=0x%08x hook=0x%08x\n",
	service, hook );
  hook_internal *hi = new hook_internal;
  hi->service = service;
  hi->hook = hook;
  hooked_services.push_back( hi );
  return (int) hook;
}

int __VMM_lstrlen( char *s ) {
  return strlen( s );
}

int VMM_GetVxDName( int laddr, char *pszName ) {
  int i, j;
  vxd_internal *vxd;

  for ( i = 0; i < vxd_list.size(); i++ ) {
	vxd = vxd_list[i];
	for ( j = 0; j < vxd->secnum; j++ ) {
		if ( laddr > (int) vxd->sectbl[j] &&
		     laddr < (int) vxd->sectbl[j] + vxd->secsz[j] ) {
			sprintf( pszName, "%s(%02d) + %X", vxd->name,
				j, laddr - (int) vxd->sectbl[j] );
			printf( "VMM.GetVxDName laddr=0x%08x, ret=`%s`\n",
				laddr, pszName );
			return (int) vxd->sectbl[0];
		}
	}
  }
  printf( "VMM.GetVxDName laddr=0x%08x, system\n", laddr );
  sprintf( pszName, "%X\n", laddr );
  return 0;
}

// IOS subsystem

void dump_ior( IOR *ior ) {
  printf( "\t%08x\tIOR_next\n", ior->IOR_next );
  printf( "\t%04x\t\tIOR_func\n", ior->IOR_func );
  printf( "\t%04x\t\tIOR_status\n", ior->IOR_status );
  printf( "\t%08x\tIOR_flags\n", ior->IOR_flags );
  printf( "\t%08x\tIOR_callback\n", ior->IOR_callback );
  printf( "\t%08x\tIOR_start_addr[0]\n", ior->IOR_start_addr[0] );
  printf( "\t%08x\tIOR_start_addr[1]\n", ior->IOR_start_addr[1] );
  printf( "\t%08x\tIOR_xfer_count\n", ior->IOR_xfer_count );
  printf( "\t%08x\tIOR_buffer_ptr\n", ior->IOR_buffer_ptr );
  printf( "\t%08x\tIOR_private_client\n", ior->IOR_private_client );
  printf( "\t%08x\tIOR_private_IOS\n", ior->IOR_private_IOS );
  printf( "\t%08x\tIOR_private_port\n", ior->IOR_private_port );
  printf( "\t%04x\t\t_ureq.sdeffsd_req_usage._IOR_ioctl_drive\n",
	 ior->_ureq.sdeffsd_req_usage._IOR_ioctl_drive );
  printf( "\t%04x\t\t_ureq.sdeffsd_req_usage._IOR_ioctl_function\n",
	 ior->_ureq.sdeffsd_req_usage._IOR_ioctl_function );
  printf( "\t%08x\t_ureq.sdeffsd_req_usage._IOR_ioctl_control_param\n",
	 ior->_ureq.sdeffsd_req_usage._IOR_ioctl_control_param );
  printf( "\t%08x\t_ureq.sdeffsd_req_usage._IOR_ioctl_buffer_ptr\n",
	 ior->_ureq.sdeffsd_req_usage._IOR_ioctl_buffer_ptr );
  printf( "\t%08x\t_ureq.sdeffsd_req_usage._IOR_ioctl_client_params\n",
	 ior->_ureq.sdeffsd_req_usage._IOR_ioctl_client_params );
  printf( "\t%08x\t_ureq.sdeffsd_req_usage._IOR_ioctl_return\n",
	 ior->_ureq.sdeffsd_req_usage._IOR_ioctl_return );
  printf( "\t%08x\tIOR_req_req_handle\n", ior->IOR_req_req_handle );
  printf( "\t%08x\tIOR_req_vol_handle\n", ior->IOR_req_vol_handle );
  printf( "\t%08x\tIOR_sgd_lin_phys\n", ior->IOR_sgd_lin_phys );
  printf( "\t%02x\t\tIOR_num_sgds\n", ior->IOR_num_sgds );
  printf( "\t%02x\t\tIOR_vol_designtr\n", ior->IOR_vol_designtr );
  printf( "\t%04x\t\tIOR_ios_private_1\n", ior->IOR_ios_private_1 );
  printf( "\t%08x\tIOR_reserved_2[0]\n", ior->IOR_reserved_2[0] );
  printf( "\t%08x\tIOR_reserved_2[1]\n", ior->IOR_reserved_2[1] );
}

void default_compl_handler( IOP *iop ) {
  puts( "default request completion handler called" );
}

void __IOS_SendCommand( IOR *ior, DCB *dcb ) {
  char *cmdname;
  IOP *iop;

//  dump_ior( ior );
  switch ( ior->IOR_func ) {
  case 0: cmdname = "read"; break;
  case 1: cmdname = "write"; break;
  case 2: cmdname = "verify"; break;
  case 3: cmdname = "cancel"; break;
  case 4: cmdname = "writev"; break;
  case 5: cmdname = "media_check"; break;
  case 6: cmdname = "media_check_reset"; break;
  case 7: cmdname = "load_media"; break;
  case 8: cmdname = "eject_media"; break;
  case 9: cmdname = "lock_media"; break;
  case 0x0a: cmdname = "unlock_media"; break;
  case 0x0b: cmdname = "request_sense"; break;
  case 0x0c: cmdname = "compute_geom"; break;
  case 0x0d: cmdname = "gen_ioctl"; break;
  case 0x0e: cmdname = "format"; break;
  case 0x0f: cmdname = "scsi_pass_through"; break;
  case 0x10: cmdname = "clear_queue"; break;
  case 0x11: cmdname = "dos_reset"; break;
  case 0x12: cmdname = "scsi_request"; break;
  case 0x22: cmdname = "set_write_status"; break;
  case 0x23: cmdname = "restart_queue"; break;
  case 0x24: cmdname = "abort_queue"; break;
  case 0x25: cmdname = "spin_down"; break;
  case 0x26: cmdname = "spin_up"; break;
  case 0x27: cmdname = "flush_drive"; break;
  case 0x28: cmdname = "flush_drive_and_discard"; break;
  case 0x29: cmdname = "fsd_ext"; break;
  default: cmdname = "unknown";
  }

  printf( "IOS.SendCommand cmd='%s'", cmdname );
  if ( ior->IOR_flags & 0x8000 ) printf( " IORF_SRB_VALID" );
  if ( ior->IOR_flags & 0x2 ) printf( " IORF_SCATTER_GATHER" );
  if ( ior->IOR_flags & 0x400 ) printf( " IORF_VERSION_002" );
  if ( ior->IOR_flags & 0x100 )
	printf( " IORF_SYNC_COMMAND" );
  else
	printf( " callback=0x%08x", (int)ior->IOR_callback );
  putchar( '\n' );
  
  switch ( ior->IOR_func ) {
  case 0:
	printf( "IOS.SendCommand buf=0x%08x count=0x%08x\n",
		ior->IOR_buffer_ptr, ior->IOR_xfer_count );
	memset( (void*) ior->IOR_buffer_ptr, 0, ior->IOR_xfer_count );
	break;
  case 0x0d:
	break;
  }

// should be iop be malloc'ed (driver wants to free it at some point)
  iop = (IOP*) ((char*)ior - sizeof(IOP) + sizeof(IOR));
  IOP_callback_entry *iop_cb = &iop->IOP_callback_table[0];
  iop_cb->IOP_CB_address = default_compl_handler;

  iop->IOP_callback_ptr = &iop->IOP_callback_table[1];

  iop->IOP_calldown_ptr = dcb->DCB_cmn.DCB_ptr_cd;

//  puts( "internal debug on" );
//  char buf[256];
//  VMM_GetVxDName( (int) iop->IOP_calldown_ptr->DCB_cd_io_address, buf );
//  printf( "SendCommand: passing iop to %s\n", buf );
//  puts( "internal debug off" );

  iop->IOP_calldown_ptr->DCB_cd_io_address( iop );
}

extern "C" {
void __ILB_service_rtn( ISP *pisp );
void __ILB_dprintf_rtn();
void __ILB_wait_10th_sec();
void __ILB_internal_request( IOP *iop, DCB *dcb, void (*vector)(IOP*) );
unsigned long __ILB_io_criteria_rtn( void *ior );
unsigned long __ILB_int_io_criteria_rtn( void *iop );
void __ILB_enqueue_iop();
void __ILB_dequeue_iop();
}

void __ILB_service_rtn( ISP *pisp ) {
  char *serv, *ptr;
  int expan_len;
  IOP *iop;
  ISP_IOP_create *isp_ioc;
  ISP_mem_alloc *isp_mma;
  ISP_mem_dealloc *isp_mmd;
  ISP_calldown_insert *isp_cdi;
  ISP_pick_drive_letter *isp_pdl;
  DCB_cd_entry *cb_ent;

  pisp->ISP_result = 1;
  switch ( pisp->ISP_func ) {
  case 0: serv = "create ddb"; break;
  case 1: serv = "create dcb (physical)"; break;
  case 2:
	serv = "create iop";
	isp_ioc = (ISP_IOP_create*) pisp;
	ptr = (char*) malloc( isp_ioc->ISP_IOP_size );
	memset( ptr, 0, isp_ioc->ISP_IOP_size );
	isp_ioc->ISP_IOP_ptr = (unsigned int) ptr;
	printf( "ILB service routine: this=0x%08x, IOP=0x%08x, size=%d, delta=0x%x\n",
		__ILB_service_rtn, isp_ioc->ISP_IOP_ptr, isp_ioc->ISP_IOP_size, isp_ioc->ISP_delta_to_ior );
	iop = (IOP*) (ptr + isp_ioc->ISP_delta_to_ior - (int)&((IOP*)0)->IOP_ior);
	iop->IOP_physical = iop;
	iop->IOP_physical_dcb = (DCB_COMMON*) cdr_dcb_ph;
	iop->IOP_original_dcb = (DCB_COMMON*) cdr_dcb_ph;
	pisp->ISP_result = 0;
	break;
  case 3:
	serv = "allocate memory";
	isp_mma = (ISP_mem_alloc*) pisp;
	isp_mma->ISP_mem_ptr = (unsigned long) malloc( isp_mma->ISP_mem_size );
	printf( "ILB service routine: alloc size = %d, ptr = 0x%08x\n",
		isp_mma->ISP_mem_size, isp_mma->ISP_mem_ptr );
	memset( (void*) isp_mma->ISP_mem_ptr, 0, isp_mma->ISP_mem_size );
	pisp->ISP_result = 0;
	break;
  case 4:
	serv = "deallocate memory";
	isp_mmd = (ISP_mem_dealloc*) pisp;
	printf( "ILB service routine: ptr = 0x%08x\n",
		 isp_mmd->ISP_mem_ptr_da );
	free( (void*) isp_mmd->ISP_mem_ptr_da );
	pisp->ISP_result = 0;
	break;
  case 5:
  	serv = "insert entry into calldown table";
	isp_cdi = (ISP_calldown_insert*) pisp;
	printf( "ILB service routine: request routine ptr = 0x%08x, expan_len=%d\n",
		((char*) isp_cdi->ISP_i_cd_req) - vxd->sectbl[0],
		isp_cdi->ISP_i_cd_expan_len );
	expan_len = isp_cdi->ISP_i_cd_dcb->DCB_expansion_length +
		isp_cdi->ISP_i_cd_expan_len;
	cb_ent = new DCB_cd_entry;
	memset( cb_ent, 0, sizeof(DCB_cd_entry) );
	cb_ent->DCB_cd_io_address = (void (*)(IOP*)) isp_cdi->ISP_i_cd_req;
	cb_ent->DCB_cd_expan_off = -expan_len;
	cb_ent->DCB_cd_next = isp_cdi->ISP_i_cd_dcb->DCB_ptr_cd;
	isp_cdi->ISP_i_cd_dcb->DCB_expansion_length = expan_len;
	isp_cdi->ISP_i_cd_dcb->DCB_ptr_cd = cb_ent;
	pisp->ISP_result = 0;
//	puts( "internal debug on" );
//	VMM_GetVxDName( (int) isp_cdi->ISP_i_cd_dcb->DCB_ptr_cd->DCB_cd_io_address, buf );
//	printf( "DCB_cd_ptr points to %s\n", buf );
//	puts( "internal debug off" );
  	break;
  case 6: serv = "associate dcb and relative drive"; break;
  case 7: serv = "return dcb for logical volume"; break;
  case 8: serv = "get first/next dcb"; break;
  case 9: serv = "deallocate ddb"; break;
  case 10: serv = "destroy dcb"; break;
  case 11: serv = "ask for logical-to-physical info"; break;
  case 12: serv = "ask if we can remove a device"; break;
  case 13: serv = "indicate that device has been removed"; break;
  case 14: serv = "indicate that device has arrived"; break;
  case 15: serv = "disassociate dcb & drive"; break;
  case 16:
	serv = "pick a drive letter";
	isp_pdl = (ISP_pick_drive_letter*) pisp;
	isp_pdl->ISP_p_d_l_letter[0] = 'D';
	isp_pdl->ISP_p_d_l_letter[1] = ':';
	pisp->ISP_result = 0;
	break;
  case 17: serv = "read registry"; break;
  case 18: serv = "find ldm entry"; break;
  case 19: serv = "delete ldm entry"; break;
  case 20: serv = "broadcast an aep from a driver"; break;
  default: serv = "unknown";
  }
  printf( "ILB service routine: isp=0x%08x, service=`%s`\n", pisp, serv );
}

void __ILB_dprintf_rtn() {
  printf( "ILB dprintf routine\n" );
}

void __ILB_wait_10th_sec() {
  printf( "ILB Wait 10th second\n" );
}

void __ILB_internal_request( IOP *iop, DCB *dcb, void (*vector)(IOP*) ) {
  printf( "ILB internal request iop=0x%08x, dcb=0x%08x, vector=0x%08x\n",
	iop, dcb, vector );
  if ( vector ) return;
  __IOS_SendCommand( &iop->IOP_ior, dcb );
//  iop->IOP_calldown_ptr = dcb->DCB_cmn.DCB_ptr_cd;
//  iop->IOP_calldown_ptr->DCB_cd_io_address( iop );
}

unsigned long __ILB_io_criteria_rtn( void *ior ) {
  printf( "ILB io criteria routine ior=0x%08x\n", ior );
  return 0;
}

unsigned long __ILB_int_io_criteria_rtn( void *iop ) {
  printf( "ILB int io criteria routine iop=0x%08x\n", iop );
  return 0;
}

void __ILB_enqueue_iop() {
  printf( "ILB enqueue IOP\n" );
}

void __ILB_dequeue_iop() {
  printf( "ILB dequeue IOP\n" );
}

void dump_drp( int drp ) {
  char *p = (char*) drp;

  printf( "\t'%-8.8s'\teye catcher string\n", p );	p += 8;
  printf( "\t%08x\tload group number\n", *(int*)p );	p += 4;
  printf( "\t%08x\tasync event routine\n", *(int*)p );	p += 4;
  printf( "\t%08x\tILB virtual address\n", *(int*)p );	p += 4;
  printf( "\t'%-16.16s' ascii name\n", p );		p += 16;
  printf( "\t%d\t\tdriver revision\n", *p );		p++;
  printf( "\t%08x\tfeature code\n", *(int*)p );		p += 4;
  printf( "\t%04x\t\ti/f requirements\n", *(unsigned short*)p ); p += 2;
  printf( "\t%02x\t\tbus type\n", *(unsigned char*)p );	p++;
  printf( "\t%04x\t\tregistration results\n", *(unsigned short*)p ); p += 2;
  printf( "\t%08x\tfield passed in on initialize aep\n", *(int*)p ); p += 4;
  printf( "\t%04x\t\treserved1\n", *(unsigned short*)p ); p += 2;
  printf( "\t%08x\treserved2\n", *(int*)p );		p += 4;
}

void dump_ilb( int ilb ) {
  char *p = (char*) ilb;

  printf( "\t%08x\tservice routine entry point into IOS\n",
	*(unsigned int*)p ); p += 4;
  printf( "\t%08x\tdprintf routine entry point into IOS\n",
	*(unsigned int*)p ); p += 4;
  printf( "\t%08x\tcpu spin wait for 10th second\n",
	*(unsigned int*)p ); p += 4;
  printf( "\t%08x\tvector for issuing internal requests called by dragon\n"
	  "\t\tdrivers (drivers in DCB calldown stack)\n",
	*(unsigned int*)p ); p += 4;
  printf( "\t%08x\troutine to satisfy IOR criteria called by dragon clients\n"
	  "\t\twith IOS allocated IOR\n",
	*(unsigned int*)p ); p += 4;
  printf( "\t%08x\troutine to satisfy IOP criteria called by dragon drivers\n"
	  "\t\t(drivers in DCB calldown stack)\n",
	*(unsigned int*)p ); p += 4;
  printf( "\t%08x\t32 bit offset of this drivers dvt\n",
	*(unsigned int*)p ); p += 4;
  printf( "\t%08x\tpointer to IOS memory pool\n",
	*(unsigned int*)p ); p += 4;
  printf( "\t%08x\tIOP enqueue routine\n",
	*(unsigned int*)p ); p += 4;
  printf( "\t%08x\tIOP dequeue routine\n",
	*(unsigned int*)p ); p += 4;
  printf( "\t%08x\treserved1 (must be zero)\n",
	*(unsigned int*)p ); p += 4;
  printf( "\t%08x\treserved2 (must be zero)\n",
	*(unsigned int*)p ); p += 4;
  printf( "\t%04x\tflags\n",
	*(unsigned int*)p ); p += 2;
  printf( "\t%02x\t\tthe number of times driver has been called with and\n"
	  "\t\tinitialize aep (ilb_reserved_43)\n",
	*(unsigned int*)p ); p++;
  printf( "\t%02x\t\treserved3 (must be zero)\n",
	*(unsigned int*)p ); p++;
}

extern "C" {
void ILB_service_rtn();
void ILB_dprintf_rtn();
void ILB_wait_10th_sec();
void ILB_internal_request();
void ILB_io_criteria_rtn();
void ILB_int_io_criteria_rtn();
void ILB_enqueue_iop();
void ILB_dequeue_iop();
}

void IOS_Register( int pdrp ) {
  char *p = (char*) pdrp;
  char *ilb = *(char**)&p[16];
  char *ida;	// IOS Data Area
  int lgn = *(int*)&p[8];

  printf( "IOS.Register (pdrp=0x%08x) lgn=0x%08x\n",
	pdrp - (int) vxd->sectbl[1], lgn );
//  dump_drp( pdrp );

  ida = (char*) malloc( 0x6e );
  memset( ida, 0, 0x58 );

//  for ( int i = 0; i < 52; i++ ) ilb[i] = i + 0x12;
  memset( ilb, 0, 52 );
  *(void**)ilb = (void*) ILB_service_rtn; ilb += 4;
  *(void**)ilb = (void*) ILB_dprintf_rtn; ilb += 4;
  *(void**)ilb = (void*) ILB_wait_10th_sec; ilb += 4;
  *(void**)ilb = (void*) ILB_internal_request; ilb += 4;
  *(void**)ilb = (void*) ILB_io_criteria_rtn; ilb += 4;
  *(void**)ilb = (void*) ILB_int_io_criteria_rtn; ilb += 4;
  ilb += 4;
  *(char**)ilb = (char*) ida; ilb += 4;
  *(void**)ilb = (void*) ILB_enqueue_iop; ilb += 4;
  *(void**)ilb = (void*) ILB_dequeue_iop; ilb += 4;
  
//  puts( "ILB contents:" );
//  dump_ilb( *(int*)&p[16] );
  vxd->drp = p;
  *(int*)&p[44] = 1; // DPR_REMAIN_RESIDENT
}

void dump_srb( SCSI_REQUEST_BLOCK *srb ) {
  printf( "\t%04x\t\tLength\n", srb->Length );
  printf( "\t%02x\t\tFunction\n", srb->Function );
  printf( "\t%02x\t\tSrbStatus\n", srb->SrbStatus );
  printf( "\t%02x\t\tScsiStatus\n", srb->ScsiStatus );
  printf( "\t%02x\t\tPathId\n", srb->PathId );
  printf( "\t%02x\t\tTargetId\n", srb->TargetId );
  printf( "\t%02x\t\tLun\n", srb->Lun );
  printf( "\t%02x\t\tQueueTag\n", srb->QueueTag );
  printf( "\t%02x\t\tQueueAction\n", srb->QueueAction );
  printf( "\t%02x\t\tCdbLength\n", srb->CdbLength );
  printf( "\t%02x\t\tSenseInfoBufferLength\n", srb->SenseInfoBufferLength );
  printf( "\t%08x\tSrbFlags\n", srb->SrbFlags );
  printf( "\t%08x\tDataTransferLength\n", srb->DataTransferLength );
  printf( "\t%08x\tTimeOutValue\n", srb->TimeOutValue );
  printf( "\t%08x\tDataBuffer\n", srb->DataBuffer );
  printf( "\t%08x\tSenseInfoBuffer\n", srb->SenseInfoBuffer );
  printf( "\t%08x\tNextSrb\n", srb->NextSrb );
  printf( "\t%08x\tOriginalRequest\n", srb->NextSrb );
  printf( "\t%08x\tSrbExtension\n", srb->NextSrb );
  printf( "\t%08x\tQueueSortKey\n", srb->NextSrb );
  putchar( '\t' );
  for ( int i = 0; i < 16; i++ ) printf( "%02X ", srb->Cdb[i] );
  printf( " Cdb\n" );
}

void unhandled_requests_tracker( IOP *iop ) {
  SCSI_REQUEST_BLOCK *srb;

  printf( "unhandled request iop=0x%08x", iop );
  if ( iop->IOP_ior.IOR_flags & 0x8000 ) {	// IORF_SRB_VALID
	printf( " srb=0x%08x\n", iop->IOP_srb );
	srb = (SCSI_REQUEST_BLOCK*) iop->IOP_srb;
	dump_srb( srb );
// we can reply to 5A 08 01 00 00 00 00 00 20 00 00 00
	if ( srb->CdbLength == 0xc &&
	  !memcmp( srb->Cdb, "\x5A\x08\x01\x00\x00\x00\x00\x00\x20\x00\x00\x00", 0xc ) ) {
		memcpy( srb->DataBuffer, "\x00\x12\x21\x00\x00\x00\x00\x00\x01\x0a\x80\x05\x00\x00\x00\x00\x03\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00", 0x20 );
	}
	printf( "scsi request completed callback=0x%08x\n",
		(int) iop->IOP_ior.IOR_callback - (int) vxd->sectbl[0] );
	((SCSI_REQUEST_BLOCK*)iop->IOP_srb)->SrbStatus = 1; // SRB_STATUS_SUCCESS
	iop->IOP_ior.IOR_callback( iop );
  }
  else {
	putchar( '\n' );
	dump_ior( &iop->IOP_ior );
  }
}

void create_cdr_dcb() {
  int *dev_node;
  dev_node = (int*) malloc( sizeof(int) );
  *dev_node = 0x99;

  cdr_dcb_ph = new DCB_cdrom;
  memset( cdr_dcb_ph, 0, sizeof(DCB_cdrom) );	// with 0x20 sense fails
  cdr_dcb_ph->DCB_filler_2.DCB_cmn.DCB_expansion_length = 0;
  cdr_dcb_ph->DCB_filler_2.DCB_srb_ext_size = 0;
  cdr_dcb_ph->DCB_cd_current_block_size = 0x800;

// still not known magics
  *((char*)cdr_dcb_ph->DCB_TOC + 3) = 1;
  *(int*)((char*)cdr_dcb_ph->DCB_TOC + 4) = 0x00400;
  *(int*)((char*)cdr_dcb_ph->DCB_TOC + 8) = 0x20000;

  cdr_dcb_ph->DCB_filler_2.DCB_cmn.DCB_ptr_cd = 0; // will be set later
  cdr_dcb_ph->DCB_filler_2.DCB_cmn.DCB_physical_dcb = (unsigned int) cdr_dcb_ph;
  cdr_dcb_ph->DCB_filler_2.DCB_cmn.DCB_device_type = 5;	// cdrom
// flags: DCB_DEV_PHYSICAL  
  cdr_dcb_ph->DCB_filler_2.DCB_cmn.DCB_device_flags = 0xd021d004;
// flags2: DCB_DEV2_ATAPI_DEVICE
//  cdr_dcb_ph->DCB_filler_2.DCB_cmn.DCB_device_flags2 = 2;
  cdr_dcb_ph->DCB_filler_2.DCB_cmn.DCB_dmd_flags = 0x00070800;	// for cdtsd
  cdr_dcb_ph->DCB_filler_2.DCB_cmn.DCB_drive_lttr_equiv = 3;	/* D: */
  cdr_dcb_ph->DCB_filler_2.DCB_max_xfer_len = 512;
// for cdvsd
  cdr_dcb_ph->DCB_filler_2.DCB_bus_type = 1;
  cdr_dcb_ph->DCB_filler_2.DCB_dev_node = dev_node;
  memcpy( cdr_dcb_ph->DCB_filler_2.DCB_inquiry_flags,
	"\x05\x80\x02\x32\x5b\x00\x00\x00", 8 );

#if 0
  cdr_dcb_lg = new DCB_cdrom;
  memcpy( cdr_dcb_lg, cdr_dcb_ph, sizeof(DCB_cdrom) );
// flags: DCB_DEV_LOGICAL
  cdr_dcb_lg->DCB_filler_2.DCB_cmn.DCB_device_flags = 0xd0214004;
  cdr_dcb_ph->DCB_filler_2.DCB_cmn.DCB_drive_lttr_equiv = 3;	/* D: */
#endif

  DCB_cd_entry *cb_ent = new DCB_cd_entry;
  memset( cb_ent, 0, sizeof(DCB_cd_entry) ); // 0x43
  cb_ent->DCB_cd_io_address = unhandled_requests_tracker;
  cb_ent->DCB_cd_expan_off = 0;
  cb_ent->DCB_cd_next = 0;
  cdr_dcb_ph->DCB_filler_2.DCB_cmn.DCB_ptr_cd = cb_ent;

#if 0
  DCB_cd_entry *cb_ent_lg = new DCB_cd_entry;
  memcpy( cb_ent_lg, cb_ent, sizeof(DCB_cd_entry) );
  cdr_dcb_lg->DCB_filler_2.DCB_cmn.DCB_ptr_cd = cb_ent_lg;
#endif
}

void aep_init() {
  AEP_bi_init init;
  AEP_boot_done bdone;
  AEP_dcb_config dcfg;
  int i;

/*
  printf( "aep_disp_tbl offs=0x%08x\n",
	 *(int*)(sectbl[0] + 0x22b24) - (int)sectbl[0] );
  printf( "aep_disp_tbl[0]=0x%08x\n",
	 *(int*)(sectbl[0] + 0x231c6) - (int)sectbl[0] );
  printf( "aep_disp_tbl[1]=0x%08x\n",
	 *(int*)(sectbl[0] + 0x231c6 + 4) - (int)sectbl[0] );
  printf( "aep_disp_tbl[2]=0x%08x\n",
	 *(int*)(sectbl[0] + 0x231c6 + 2 * 4) - (int)sectbl[0] );
  printf( "aep_disp_tbl[3]=0x%08x\n",
	 *(int*)(sectbl[0] + 0x231c6 + 3 * 4) - (int)sectbl[0] );
  printf( "aep_disp_tbl[4]=0x%08x\n",
	 *(int*)(sectbl[0] + 0x231c6 + 4 * 4) - (int)sectbl[0] );
*/
  printf( "AEP_INITIALIZE in progress\n" );
  for ( i = 0; i < vxd_list.size(); i++ ) {
  	vxd = vxd_list[i];
	if ( !vxd->drp ) continue;
  	printf( "async ev handler = 0x%08x\n",
	  *(unsigned int*)&vxd->drp[12] - (int)vxd->sectbl[0] );
/*
	if ( i == 0 ) {
		int *tbl = *(int**)(vxd->sectbl[0] + 0x1470b );
		printf( "vector table at 0x%08x\n", tbl );
		for ( int j = 0; j < 20; j++ )
			printf( "\thandler[%d]=0x%08x\n", j, tbl[j] );
	}
*/
	init.AEP_bi_i_hdr.AEP_func = 0;	// AEP_INITIALIZE
	init.AEP_bi_reference_data = *(int*)&vxd->drp[0x2e];
	asm ( "pusha\n\t" );
	((void (*)(AEP_bi_init*)) *(int*)&vxd->drp[12])(&init);
	asm ( "popa\n\t" );
  }
  printf( "AEP_CONFIG_DCB in progress\n" );
  dcfg.AEP_d_c_hdr.AEP_func = 3;	// AEP_CONFIG_DCB
  for ( i = 0; i < vxd_list.size(); i++ ) {
  	vxd = vxd_list[i];
	if ( !vxd->drp ) continue;
	dcfg.AEP_d_c_dcb = (DCB_COMMON*) cdr_dcb_ph;
	asm ( "pusha\n\t" );
	((void (*)(AEP_dcb_config*)) *(int*)&vxd->drp[12])(&dcfg);
	asm ( "popa\n\t" );
#if 0
	dcfg.AEP_d_c_dcb = (DCB_COMMON*) cdr_dcb_lg;
	asm ( "pusha\n\t" );
	((void (*)(AEP_dcb_config*)) *(int*)&vxd->drp[12])(&dcfg);
	asm ( "popa\n\t" );
#endif
  }
  printf( "AEP_BOOT_COMPLETE in progress\n" );
  bdone.AEP_b_d_hdr.AEP_func = 2;	// AEP_BOOT_COMPLETE
  for ( i = 0; i < vxd_list.size(); i++ ) {
  	vxd = vxd_list[i];
	if ( !vxd->drp ) continue;
	asm ( "pusha\n\t" );
	((void (*)(AEP_boot_done*)) *(int*)&vxd_list[i]->drp[12])(&bdone);
	asm ( "popa\n\t" );
  }
}

// Configuration Manager
int CONFIGMG_Get_Parent( int *pParentNode, int node, int flags ) {
  printf( "CONFIGMG.Get_Parent node=0x%08x flags=0x%08x\n",
	node, flags );
  *pParentNode = 1;
  return 0; // CR_SUCCESS
}

int CONFIGMG_Read_Registry_Value( int node, char *szSubkey, char *szValueName,
	int expectedType, void *buf, int pLen, int flags ) {
//  printf( "CONFIGMG.ReadRegistryValue 1: node=0x%08x subkey=%s\n",
//	node, szSubkey );
//  printf( "CONFIGMG.ReadRegistryValue 2: type=0x%08x buf=0x%08x flags=0x%08x\n",
//	expectedType, buf, flags );
  printf( "CONFIGMG.ReadRegistryValue valuename='%s'\n",
	szValueName );
  return 0x25; // CR_NO_SUCH_VALUE
}

int CONFIGMG_Write_Registry_Value( int node, char *szSubkey, char *szValueName,
	int dwType, void *buf, int pLen, int flags ) {
//  printf( "CONFIGMG.Write_Registry_Value 1: node=0x%08x subkey=%s\n"
//	node, szSubkey );
//  printf( "CONFIGMG.Write_Registry_Value 2: type=0x%08x flags=0x%08x\n",
//	dwType, flags );
  printf( "CONFIGMG.Write_Registry_Value valuename='%s'\n", szValueName );
  printf( "CONFIGMG.Write_Registry_Value value='%s'\n", buf );
  return 0; // CR_SUCCESS
}

struct callback1_struct {
  void* (*handler)(int);
  unsigned int refdata;
};

vector <callback1_struct> appy_callbacks;

// Shell functions
int SHELL_CallAtAppyTime( void* (*handler)(int), unsigned int refdata,
	unsigned int flags ) {
  callback1_struct cb;

  printf( "SHELL.CallAtAppyTime (handler=0x%08x, refdata=0x%08x, flags=0x%x)\n",
  	handler, refdata, flags );
  cb.handler = handler;
  cb.refdata = refdata;
  appy_callbacks.push_back( cb );

  return 0; // CR_SUCCESS
}

// IFSMgr subsystem

int IFSMGR_RegisterMount( int pfunc, int version, int type ) {
  printf( "IFSMGR.RegisterMount (pfunc=0x%08x, version=%d, type=%d)\n",
	pfunc, version, type );
  vxd->pIfs = (int (*)(ioreq*)) pfunc;
  return 4;	// provider id
}

int IFSMGR_UniToBCS( unsigned char *pBCSStr, unsigned short *pUniStr,
	unsigned int length, unsigned int maxLength, int charSet ) {
  int i;

#if 0
  printf( "IFSMGR_UniToBCS cstr=0x%08x str=0x%08x len=%d\n",
  	pBCSStr, pUniStr, length );

  if ( uni_verbose ) {
	printf( "IFSMGR string=`" );
	for ( i = 0; i < length && i < maxLength; i++ ) putchar( pUniStr[i] );
	printf( "`\n" );
  }
#endif 

//  iconv_t id = iconv_open( "koi8-r", "ucs-2be" );
//  if ( id == (iconv_t)-1 ) {
//  	puts( "IFSMGR_UniToBCS unable to open convertion table" );
	length /= 2;
	maxLength /= 2;
  	for ( i = 0; i < length && i < maxLength; i++ ) pBCSStr[i] = pUniStr[i];
	return i;
//  }
#if 0
  unsigned short *p = pUniStr;
  unsigned char *d = pBCSStr;
  int len;
  
  iconv( id, (char**) &pUniStr, &length, (char**) &pBCSStr, &maxLength );
  len = (char*)pUniStr - (char*)p;
  if ( len > 2 ) {
	printf( "IFSMGR_UniToBCS translated string: "  );
	for ( i = 0; i < len; i++ ) putchar( d[i] );
	putchar( '\n' );
  }
  return len;
#endif 
}

int IFSMGR_UniToUpper( unsigned short *pUniUpStr,
	unsigned short *pUniStr, int len ) {
  int i;
  len /= 2;
  
//  printf( "IFSMGR_UniToUpper upstr=0x%08x str=0x%08x len=%d\n",
//  	pUniUpStr, pUniStr, len );

  if ( uni_verbose ) {
	printf( "IFSMGR string=`" );
	for ( i = 0; i < len; i++ ) putchar( pUniStr[i] );
	printf( "`\n" );
  }

  for ( i = 0; i < len; i++ )
	pUniUpStr[i] = toupper( pUniStr[i] );
  return i;
}

int IFSMGR_BCSToUni( unsigned short *pUniStr, unsigned char *pBCSStr,
	unsigned int length, int charSet ) {
  int i;

  puts( "IFSMGR.BCSToUni str: " );
  for ( i = 0; i < length; i++ ) {
  	putchar( pBCSStr[i] );
  	pUniStr[i] = pBCSStr[i];
  }
  putchar( '\n' );
  return i;
}

int IFSMGR_CheckAccessConflict( int fSoftCompatibleDisable, void *pfmode,
	int (*pfn)(void*, const void*), const void* FSDCookie ) {
  puts( "IFSMGR.CheckAccessConflict" );
  return 0;
}

void ifs_mount_request( vxd_internal *vxd ) {
  int rc;

  printf( "IR_FSD_MOUNT (pIfs=0x%08x) in progress\n",
	(int) vxd->pIfs - (int) vxd->sectbl[0] );
  vrp = (VRP*) malloc( sizeof(VRP) );
  memset( vrp, 0, sizeof(VRP) );
//  for ( int i = 0; i < sizeof(VRP); i++ )
//	((char*)vrp)[i] = 0x52 + i;
  vrp->VRP_max_sgd = 0;
  vrp->VRP_max_req_size = sizeof(IOP);
  vrp->VRP_delta_to_ior = (int) &((IOP*)0)->IOP_ior;
  vrp->VRP_partition_offset = (int) cdr_dcb_ph;
  ((DCB_COMMON*)cdr_dcb_ph)->DCB_vrp_ptr = vrp;
//  ((DCB_COMMON*)cdr_dcb_lg)->DCB_vrp_ptr = (unsigned int) vrp;
  ioreq *ir = (ioreq*) malloc( sizeof(ioreq) );
  memset( ir, 0, sizeof(ioreq) );
  ir->ir_flags = 0;		// IR_FSD_MOUNT
  ir->ir_volh = (unsigned int) vrp;
  ir->ir_mntdrv = 3;		// D:
  rc = vxd->pIfs( ir );
  printf( "IR_FSD_MOUNT rc=%d ir_error=%d\n", rc, ir->ir_error );
// 608 MOUNT fsd doesn't recognize media file system
  if ( ir->ir_error ) return;
  if ( !ir->ir_vfunc ) {
	puts( "volfunc is not defined" );
	return;
  }
  if ( !ir->ir_vfunc->vfn_size ) return;
  vfs_vfunc = ir->ir_vfunc;
  vfs_rh = ir->ir_rh;
}

// VxD dispatching support

int install_vxdcall_handler( int type, int pid, int ip, void (*hand)() ) {
  int call=0xe890, jmp=0xe990;
  int disp_hand = (int) ((char*)hand - (ip + 2) - 4);
  int rc;

  if ( type & 0x8000 )
	rc = ptrace( (__ptrace_request) PTRACE_POKETEXT, pid, ip, jmp );
  else
	rc = ptrace( (__ptrace_request) PTRACE_POKETEXT, pid, ip, call );
  if ( rc == -1 ) {
	perror( "ptrace poketext call" );
	return 0;
  }
  ip += 2;
  rc = ptrace( (__ptrace_request) PTRACE_POKETEXT, pid, ip, disp_hand );
  if ( rc == -1 ) {
	perror( "ptrace poketext disp_hand" );
	return 0;
  }
  return 1;
}

int try_install_hook( int service, int pid, int ip ) {
  list <hook_internal*>::iterator it;

  for ( it = hooked_services.begin(); it != hooked_services.end(); ++it ) {
	if ( (*it)->service == (service & 0xffff7fff) ) {
		printf( "installing hooked service call=0x%08x eip=0x%08x\n",
			service, ip );
		install_vxdcall_handler( service, pid, ip, (*it)->hook );
		return 1;
	}
  }

  return 0;
}

const char *vxdcall_type( int ct ) {
  return (ct & 0x8000)? "VxDjmp": "VxDcall";
}

int vxdjmp_skip( int pid, user_regs_struct *regs ) {
  int rc;

  rc = ptrace( (__ptrace_request) PTRACE_PEEKDATA, pid, regs->esp );
  regs->esp += 4;
  regs->eip = rc;
  rc = ptrace( (__ptrace_request) PTRACE_SETREGS, pid, 0, regs );
  if ( rc == -1 ) {
	perror( "vxdjmp_skip: ptrace setregs" );
	return 1;
  }
  return 0;
}

void watch_var( int pid, user_regs_struct *regs ) {
  int rc;

#if 0
  if ( regs->eip == (int) vxd->sectbl[0] + 0x125ea ) {
	  printf( "0x125ea eax = 0x%08x, ebx = 0x%08x\n",
	  	regs->eax, regs->ebx );
	  sleep(1);
  }
  else if ( regs->eip == (int) vxd->sectbl[0] + 0x1596b ) {
	  printf( "0x1596b edi = 0x%08x, eax = 0x%08x\n",
	  	regs->edi, regs->eax );
	  sleep(1);
  }
#endif 
//  rc = ptrace( (__ptrace_request) PTRACE_PEEKDATA, pid, &ir0->ir_error );
//  if ( (rc & 0xffff) == 5 ) {
//  	kill( getpid(), SIGSEGV );
//  }
}

void vxd_breakpoint( int pid, user_regs_struct *regs ) {
//  if ( regs->eip == (int) vxd_list[0]->sectbl[0] + 0x12191 )
//  	kill( pid, SIGSEGV );
}

void log_trace( int ip ) {
  static int vxdn = 0, secn = 0,
	start = (int) vxd_list[vxdn]->sectbl[secn],
	lim = start + vxd_list[vxdn]->secsz[secn];
  static char *name = vxd_list[vxdn]->name;
  static int internal_code = 0;
  int i, j;

  if ( ip > start && ip < lim ) {
	fprintf( stderr, "%s:%d:%08x\n", name, secn, ip - start );
	if ( internal_code ) internal_code = 0;
	return;
  }

  vxd_internal *vxd;
  for ( i = 0; i < vxd_list.size(); i++ ) {
	vxd = vxd_list[i];
	for ( j = 0; j < vxd->secnum; j++ )
		if ( ip > (int) vxd->sectbl[j] &&
		     ip < (int) vxd->sectbl[j] + vxd->secsz[j] ) {

			vxdn = i;
			secn = j;
			start = (int) vxd->sectbl[j];
			lim = start + vxd->secsz[j];
			name = vxd_list[vxdn]->name;
			fprintf( stderr, "%s:%d:%08x\n", name, secn, ip - start );
			if ( internal_code ) internal_code = 0;
			return;
		}
  }
  if ( !internal_code ) {
	fprintf( stderr, "int:int:%08x\n", ip );
	internal_code = 1;
  }
}

int vxdvm_start( bool trace=false ) {
  pid_t child;
  int i, j, status, rc;
  user_regs_struct regs;

  child = fork();
  if ( !child) {
	kill( getpid(), SIGSTOP );
// control dispatcher init call
	for ( i = 0; i < vxd_list.size(); i++ ) {
		vxd = vxd_list[i];
		call_ctl_disp( vxd_list[i]->ctl_disp, 0x1b );
	}
	create_cdr_dcb();
	scsiport_init( (DCB*) cdr_dcb_ph );
// asynchronious event procedure (aep): initialization
	aep_init();
	for ( i = 0; i < vxd_list.size(); i++ ) {
		vxd = vxd_list[i];
		if ( vxd_list[i]->pIfs ) {
			ifs_mount_request( vxd_list[i] );
//			call_debug_handler();

// this stuff should be done after all drivers had initialized
			if ( appy_callbacks.size() ) {
				puts( "SHELL calling registered appy callbacks" );
				for ( j = 0; j < appy_callbacks.size(); j++ )
					appy_callbacks[j].handler( appy_callbacks[i].refdata );
				appy_callbacks.clear();
			}
			if ( vfs_rh ) winvfs_fuse();
		}
	}
	puts( "traced process is about to exit" );
	return 0;
  }
  rc = ptrace( (__ptrace_request) PTRACE_ATTACH, child );
  if ( rc == -1 ) {
	perror( "ptrace attach" );
	return 1;
  }
  kill( child, SIGCONT );

  for (;;) {
	rc = waitpid( child, &status, WUNTRACED );
	if ( rc == -1 ) {
		perror( "waitpid" );
		return 1;
	}
	if ( WIFEXITED(status) ) {
		puts( "process being traced is exited" );
		break;
	}
	if ( !WIFSTOPPED(status) ) {
		puts( "waitpid returned unexpected status" );
		continue;
	}
	rc = ptrace( (__ptrace_request) PTRACE_GETREGS, child, 0, &regs );
	if ( rc == -1 ) {
		perror( "ptrace getregs" );
		break;
	}
	rc = ptrace( (__ptrace_request) PTRACE_PEEKTEXT, child, regs.eip );
	if ( rc == -1 ) {
		printf( "ptrace peektext regs.eip=0x%08x\n", regs.eip );
		perror( "ptrace peektext" );
		break;
	}
#if 0
	fprintf( stderr, "%08x %08x %08x %08x %08x\n",
		regs.eip - (int)vxd_list[0]->sectbl[0],
		regs.eip - (int)vxd_list[0]->sectbl[1],
		regs.eip - (int)vxd_list[0]->sectbl[2],
		regs.eip - (int)vxd_list[0]->sectbl[3],
		regs.eip - (int)vxd_list[0]->sectbl[4] );
#else
	log_trace( regs.eip );
#endif
//	watch_var( child, &regs );
//	vxd_breakpoint( child, &regs );
	if ( (rc & 0xff) == 0xfa ) {
//		puts( "VxDDispatch: cli" );
		regs.eip++;
		rc = ptrace( (__ptrace_request) PTRACE_SETREGS, child, 0, &regs );
		if ( rc == -1 ) {
			perror( "ptrace setregs" );
			break;
		}
	}
	else if ( (rc & 0xff) == 0x9d ) {
//		puts( "VxDDispatch: popf" );
		regs.eip++;
		regs.esp += 4;
		rc = ptrace( (__ptrace_request) PTRACE_SETREGS, child, 0, &regs );
		if ( rc == -1 ) {
			perror( "ptrace setregs" );
			break;
		}
	}
	else if ( (rc & 0xffff) == 0x01cd ) {
		puts( "Debugger pop-up request" );
		regs.eip += 2;
		rc = ptrace( (__ptrace_request) PTRACE_SETREGS, child, 0, &regs );
		if ( rc == -1 ) {
			perror( "ptrace setregs" );
			break;
		}
	}
	else if ( (rc & 0xffff) == 0x20cd ) {
		rc = ptrace( (__ptrace_request) PTRACE_PEEKTEXT, child, regs.eip + 2 );
		if ( try_install_hook( rc, child, regs.eip ) )
			goto dont_skip_vxdcall;
		if ( (rc & 0xffff0000) == 0x010000 ) {
			if ( (rc & 0x7fff) == 0 ) {
				printf( "%s VMM.Get_VMM_Version\n",
					vxdcall_type(rc) );
			}
			else if ( (rc & 0x7ff) == 3 ) {
				printf( "%s VMM.Get_Sys_VM_Handle\n",
					vxdcall_type(rc) );
			}
			else if ( (rc & 0x7ff) == 0x4f ) {
				if ( install_vxdcall_handler( rc, child, regs.eip, (void (*)()) VMM_HeapAllocate ) ) goto dont_skip_vxdcall;
			}
			else if ( (rc & 0x7ff) == 0x51 ) {
				if ( install_vxdcall_handler( rc, child, regs.eip, (void (*)()) VMM_HeapFree ) ) goto dont_skip_vxdcall;
			}
			else if ( (rc & 0x7ff) == 0x53 ) {
				if ( install_vxdcall_handler( rc, child, regs.eip, (void (*)()) VMM_PageAllocate ) ) goto dont_skip_vxdcall;
			}
			else if ( (rc & 0x7ff) == 0x59 ) {
				if ( install_vxdcall_handler( rc, child, regs.eip, (void (*)()) VMM_PageGetAllocInfo ) ) goto dont_skip_vxdcall;
			}
			else if ( (rc & 0x7ff) == 0x63 ) {
				if ( install_vxdcall_handler( rc, child, regs.eip, (void (*)()) VMM_LinPageLock ) ) goto dont_skip_vxdcall;
			}
			else if ( (rc & 0x7ff) == 0x64 ) {
				printf( "%s VMM.LinPageUnlock\n",
					vxdcall_type(rc) );
			}
			else if ( (rc & 0x7ff) == 0x90 ) {
//				if ( install_vxdcall_handler( rc, child, regs.eip, (void (*)()) VMM_Hook_Device_Service ) ) goto dont_skip_vxdcall;
				__VMM_Hook_Device_Service( regs.eax, regs.esi );
			}
			else if ( (rc & 0x7ff) == 0x9b ) {
				if ( install_vxdcall_handler( rc, child, regs.eip, (void (*)()) VMM_ListCreate ) ) goto dont_skip_vxdcall;
			}
			else if ( (rc & 0x7ff) == 0x9c ) {
				if ( install_vxdcall_handler( rc, child, regs.eip, (void (*)()) VMM_ListDestroy ) ) goto dont_skip_vxdcall;
			}
			else if ( (rc & 0x7ff) == 0x9d ) {
				if ( install_vxdcall_handler( rc, child, regs.eip, (void (*)()) VMM_ListAllocate ) ) goto dont_skip_vxdcall;
			}
			else if ( (rc & 0x7ff) == 0x9e ) {
				if ( install_vxdcall_handler( rc, child, regs.eip, (void (*)()) VMM_List_Attach ) ) goto dont_skip_vxdcall;
			}
			else if ( (rc & 0x7ff) == 0xa1 ) {
				if ( install_vxdcall_handler( rc, child, regs.eip, (void (*)()) VMM_List_Remove ) ) goto dont_skip_vxdcall;
			}
			else if ( (rc & 0x7ff) == 0xa2 ) {
				if ( install_vxdcall_handler( rc, child, regs.eip, (void (*)()) VMM_ListDeallocate ) ) goto dont_skip_vxdcall;
			}
			else if ( (rc & 0x7ff) == 0xa3 ) {
				if ( install_vxdcall_handler( rc, child, regs.eip, (void (*)()) VMM_List_Get_First ) ) goto dont_skip_vxdcall;
			}
			else if ( (rc & 0x7ff) == 0xa4 ) {
				if ( install_vxdcall_handler( rc, child, regs.eip, (void (*)()) VMM_List_Get_Next ) ) goto dont_skip_vxdcall;
			}
			else if ( (rc & 0x7ff) == 0x101 ) {
				if ( install_vxdcall_handler( rc, child, regs.eip, (void (*)()) VMM_CreateMutex ) ) goto dont_skip_vxdcall;
			}
			else if ( (rc & 0x7ff) == 0x102 ) {
				if ( install_vxdcall_handler( rc, child, regs.eip, (void (*)()) VMM_DestroyMutex ) ) goto dont_skip_vxdcall;
			}
			else if ( (rc & 0x7ff) == 0x103 ) {
				printf( "%s VMM.GetMutexOwner\n",
					vxdcall_type(rc) );
			}
			else if ( (rc & 0x7fff) == 0x10a ) {
				printf( "%s VMM.Get_Sys_Thread_Handle\n",
					vxdcall_type(rc) );
			}
			else if ( (rc & 0x7fff) == 0x127 ) {
				printf( "%s VMM.SignalId\n",
					vxdcall_type(rc) );
			}
			else if ( (rc & 0x7fff) == 0x12d ) {
				if ( install_vxdcall_handler( rc, child, regs.eip, (void (*)()) VMM_Debug_Printf_Service ) ) goto dont_skip_vxdcall;
			}
			else if ( (rc & 0x7ff) == 0x12e ) {
				if ( install_vxdcall_handler( rc, child, regs.eip, (void (*)()) VMM_EnterMutex ) ) goto dont_skip_vxdcall;
			}
			else if ( (rc & 0x7ff) == 0x12f ) {
				if ( install_vxdcall_handler( rc, child, regs.eip, (void (*)()) VMM_LeaveMutex ) ) goto dont_skip_vxdcall;
			}
			else if ( (rc & 0x7fff) == 0x144 ) {
				printf( "%s VMM.Get_System_Time_Address\n",
					vxdcall_type(rc) );
			}
			else if ( (rc & 0x7ff) == 0x148 ) {
				if ( install_vxdcall_handler( rc, child, regs.eip, (void (*)()) VMM_RegOpenKey ) ) goto dont_skip_vxdcall;
			}
			else if ( (rc & 0x7fff) == 0x149 ) {
				printf( "%s VMM.RegCloseKey\n",
					vxdcall_type(rc) );
			}
			else if ( (rc & 0x7ff) == 0x14a ) {
				if ( install_vxdcall_handler( rc, child, regs.eip, (void (*)()) VMM_RegCreateKey ) ) goto dont_skip_vxdcall;
			}
			else if ( (rc & 0x7ff) == 0x14b ) {
				if ( install_vxdcall_handler( rc, child, regs.eip, (void (*)()) VMM_RegDeleteKey ) ) goto dont_skip_vxdcall;
			}
			else if ( (rc & 0x7fff) == 0x14e ) {
				printf( "%s VMM.RegSetValue\n",
					vxdcall_type(rc) );
			}
			else if ( (rc & 0x7fff) == 0x151 ) {
				if ( install_vxdcall_handler( rc, child, regs.eip, (void (*)()) VMM_RegQueryValueEx ) ) goto dont_skip_vxdcall;
			}
			else if ( (rc & 0x7fff) == 0x152 ) {
				if ( install_vxdcall_handler( rc, child, regs.eip, (void (*)()) VMM_RegSetValueEx ) ) goto dont_skip_vxdcall;
			}
			else if ( (rc & 0x7fff) == 0x16f ) {
				if ( install_vxdcall_handler( rc, child, regs.eip, (void (*)()) VMM_GetRegistryKey ) ) goto dont_skip_vxdcall;
			}
			else if ( (rc & 0x7fff) == 0x174 ) {
				if ( install_vxdcall_handler( rc, child, regs.eip, (void (*)()) VMM_Sprintf ) ) goto dont_skip_vxdcall;
			}
			else if ( (rc & 0x7fff) == 0x17c ) {
				if ( install_vxdcall_handler( rc, child, regs.eip, (void (*)()) VMM_lstrlen ) ) goto dont_skip_vxdcall;
			}
			else if ( (rc & 0x7fff) == 0x17e ) {
				if ( install_vxdcall_handler( rc, child, regs.eip, (void (*)()) VMM_GetVxDName ) ) goto dont_skip_vxdcall;
			}
			else {
				printf( "%s VMM.%04X\n",
					vxdcall_type(rc), rc & 0x7fff );
				sleep(1);
			}
		}
		else if ( (rc & 0xffff0000) == 0x100000 ) {
			if ( (rc & 0x7fff) == 0 )
				puts( "VxDCall IOS.Get_Version" );
			else if ( (rc & 0x7fff) == 4 ) {
				if ( install_vxdcall_handler( rc, child, regs.eip, (void (*)()) IOS_SendCommand ) ) goto dont_skip_vxdcall;
			}
			else if ( (rc & 0x7fff) == 7 ) {
				if ( install_vxdcall_handler( rc, child, regs.eip, (void (*)()) IOS_Register ) ) goto dont_skip_vxdcall;
			}
			else
				printf( "VxDCall IOS.%04X\n", rc & 0xffff );
		}
		else if ( (rc & 0xffff0000) == 0x400000 ) {
			if ( (rc & 0x7fff) == 0 )
				puts( "VxDCall IFSMGR.Get_Version" );
			else if ( (rc & 0x7fff) == 1 ) {
				if ( install_vxdcall_handler( rc, child, regs.eip, (void (*)()) IFSMGR_RegisterMount ) ) goto dont_skip_vxdcall;
			}
			else if ( (rc & 0x7fff) == 6 ) {
				printf( "%s IFSMGR.Get_NetTime\n",
					vxdcall_type(rc) );
			}
			else if ( (rc & 0x7fff) == 0x37 ) {
				printf( "%s IFSMGR.NetToDosTime\n",
					vxdcall_type(rc) );
			}
			else if ( (rc & 0x7fff) == 0x38 ) {
				printf( "%s IFSMGR.DosToNetTime\n",
					vxdcall_type(rc) );
			}
			else if ( (rc & 0x7fff) == 0x3b ) {
				printf( "%s IFSMGR.NetToWin32Time\n",
					vxdcall_type(rc) );
			}
			else if ( (rc & 0x7fff) == 0x3d ) {
				printf( "%s IFSMGR.MetaMatch\n",
					vxdcall_type(rc) );
			}
			else if ( (rc & 0x7fff) == 0x40 ) {
				if ( install_vxdcall_handler( rc, child, regs.eip, (void (*)()) IFSMGR_UniToBCS ) ) goto dont_skip_vxdcall;
			}
			else if ( (rc & 0x7fff) == 0x42 ) {
				if ( install_vxdcall_handler( rc, child, regs.eip, (void (*)()) IFSMGR_BCSToUni ) ) goto dont_skip_vxdcall;
			}
			else if ( (rc & 0x7fff) == 0x43 ) {
				if ( install_vxdcall_handler( rc, child, regs.eip, (void (*)()) IFSMGR_UniToUpper ) ) goto dont_skip_vxdcall;
			}
			else if ( (rc & 0x7fff) == 0x49 ) {
				printf( "%s IFSMGR.ShortToFcb\n",
					vxdcall_type(rc) );
			}
			else if ( (rc & 0x7fff) == 0x52 ) {
				if ( install_vxdcall_handler( rc, child, regs.eip, (void (*)()) IFSMGR_CheckAccessConflict ) ) goto dont_skip_vxdcall;
			}
			else if ( (rc & 0x7fff) == 0x55 ) {
				printf( "%s IFSMGR.RemoveLocks\n",
					vxdcall_type(rc) );
			}
			else if ( (rc & 0x7fff) == 0x60 ) {
				printf( "%s IFSMGR.GetTimeZoneBias\n",
					vxdcall_type(rc) );
			}
			else
				printf( "VxDCall IFSMGR.%04X\n", rc & 0xffff );
		}
		else if ( (rc & 0xffff0000) == 0x170000 ) {
			if ( (rc & 0x7fff) == 0xe ) {
				if ( install_vxdcall_handler( rc, child, regs.eip, (void (*)()) SHELL_CallAtAppyTime ) ) goto dont_skip_vxdcall;
			}
			else if ( (rc & 0x7fff) == 0x11 ) {
				printf( "%s SHELL.HookSystemBroadcast\n",
					vxdcall_type(rc) );
			}
			else
				printf( "%s SHELL.%04X\n",
					vxdcall_type(rc), rc & 0x7fff );
		}
		else if ( (rc & 0xffff0000) == 0x330000 ) {
			if ( (rc & 0x7fff) == 3 ) {
				if ( install_vxdcall_handler( rc, child, regs.eip, (void (*)()) CONFIGMG_Get_Parent ) ) goto dont_skip_vxdcall;
			}
			else if ( (rc & 0x7fff) == 0x3e ) {
				if ( install_vxdcall_handler( rc, child, regs.eip, (void (*)()) CONFIGMG_Read_Registry_Value ) ) goto dont_skip_vxdcall;
			}
			else if ( (rc & 0x7fff) == 0x3f ) {
				if ( install_vxdcall_handler( rc, child, regs.eip, (void (*)()) CONFIGMG_Write_Registry_Value ) ) goto dont_skip_vxdcall;
			}
			else
				printf( "%s CONFIGMG.%04X\n",
					vxdcall_type(rc), rc & 0x7fff );
		}
		else {
			printf( "VxDCall %04X.%04X\n", rc >> 16, rc & 0xffff );
//			puts( "internal debug on" );
//			char buf[256];
//			VMM_GetVxDName( regs.eip, buf );
//			printf( "undefined vxd request in %s\n", buf );
//			puts( "internal debug off" );
		}

		if ( rc & 0x8000 ) vxdjmp_skip( child, &regs );
		else {
			regs.eip += 6;
			rc = ptrace( (__ptrace_request) PTRACE_SETREGS, child, 0, &regs );
			if ( rc == -1 ) {
				perror( "ptrace setregs" );
				break;
			}
		}
dont_skip_vxdcall:;
	}
	if ( trace && WSTOPSIG(status) == SIGSEGV )
		rc = ptrace( (__ptrace_request) PTRACE_SINGLESTEP, child, 0, SIGSEGV );
	else {
		if ( trace )
			rc = ptrace( (__ptrace_request) PTRACE_SINGLESTEP, child, 0, 0 );
		else
			rc = ptrace( (__ptrace_request) PTRACE_CONT, child, 0, 0 );
	}
	if ( rc == -1 ) {
		perror( "ptrace singlestep" );
		break;
	}
  }
}

extern "C" {
int fuse_mount(char *mountpoint, char **argv);
void fuse_unmount(char *mountpoint);
}

int fuse_fd;

void term_handler( int ) {
  exit(0);
}

char *mountpoint;

void umount_helper() {
  fuse_unmount( mountpoint );
}

int main( int argc, char **argv ) {
#if 1
  if ( argc < 2 ) {
  	puts( "usage: winvfs <mountpoint>" );
	return 1;
  }

  mountpoint = strdup( argv[1] );
  fuse_fd = fuse_mount( mountpoint, 0 );
#endif
  
  signal( SIGTERM, term_handler );
  signal( SIGINT, term_handler );

#if 1
  vxd = load_vxd( "Udfreadr.vxd" );
  if ( !vxd ) return 1;
  
  for ( int j = 0; j < vxd->servnum; j++ )
	__VMM_Hook_Device_Service( (vxd->devnum << 16) | j, vxd->serv[j] );
  if ( vxd->servnum ) putchar( '\n' );
  vxd_list.push_back( vxd );
#else
  for ( int i = 1; i < argc; i++ ) {
	vxd = load_vxd( argv[i] );
	if ( !vxd ) break;
	for ( int j = 0; j < vxd->servnum; j++ )
		__VMM_Hook_Device_Service( (vxd->devnum << 16) | j, vxd->serv[j] );
	if ( vxd->servnum ) putchar( '\n' );
	vxd_list.push_back( vxd );
  }
#endif
  
  vxdvm_start( false );
}
